(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


Needs["FMTCodeFormatter`"];
Needs["FMTOptions`"];
Needs["FMTSymbols`"];
Needs["FMTSystem`"];


BeginPackage["FMTGeneratorAda`"];


fileNamesAda::usage="fileNamesAda[pkgName,policies] Returns a pair of file names for spec and body for the given package and translation policies.";


fmtGenerateAda::usage="fmtGenerateAda[sys,pkgName,policies] Returns, <|fileNameSpec->codeSpec,fileNameBody->codeBody|> for the given package.\n"<>
"The policies is a possibly empty association with functions for various translation steps, these policies override their respective defaults.";


Begin["Private`"];


(* Helpers *)

normalizePackageSeparatorsAda[n_String]:=StringRiffle[Capitalize/@StringSplit[n,"`"],"."]

(* default symbol translator, expects two parameters:
	- symbol or symbol name
	- name of current package (used to remove it from qualified part of the translated name)
*)
nameToStringAda[n_,currentPackageName_String]:=Module[
	{
		str,
		previousWasLowerCase=False,
		temp={}
	},
	
	If[Head[n]===Symbol,
		Switch[n,
			FMTSymbols`integerT,
				str="Integer",
			FMTSymbols`booleanT,
				str="Boolean",
			_,
				str=ToString[n]
		],
		
		str=n
	];
	
	str=StringTrim[str,currentPackageName];
	
	Scan[
		Function[c,
			If[previousWasLowerCase&&(UpperCaseQ[c]||DigitQ[c]),
				AppendTo[temp,"_"]
			];
			AppendTo[temp,c];
			previousWasLowerCase=LowerCaseQ[c]
		],
		Characters[str]
	];
	
	normalizePackageSeparatorsAda[StringJoin[temp]]
]

nameToFileRootAda[n_String]:=ToLowerCase[StringReplace[n,"."->"-"]]


(* Policy helpers, they return default policy if there is no provided by the user. *)

fileNamesFromPkgNamePolicyAda[policies_Association]:=
	Lookup[policies,
		FMTSymbols`fileNamesFromPkgName,
		Function[pkgName,
			With[
				{
					fileRoot=nameToFileRootAda[nameFromSymbolPolicyAda[policies][pkgName,""]]
				},
				
				{fileRoot<>".ads",fileRoot<>".adb"}
			]
		]
	]

nameFromSymbolPolicyAda[policies_Association]:=With[
	{
		userPolicy=Lookup[policies,FMTSymbols`nameFromSymbol,Null]
	},
	
	Function[{n,pkgName},
		If[Head[userPolicy]===Association,
			(* userPolicy is a lookup table, try there first *)
			Lookup[userPolicy,n,nameToStringAda[n,pkgName]],
			
			If[userPolicy=!=Null,
				(* userPolicy is a function (hopefully) *)
				userPolicy[n,pkgName],
				
				(* fall back to default converter *)
				nameToStringAda[n,pkgName]
			]
		]
	]
]

indentSizePolicyAda[policies_Association]:=
	Lookup[policies,
		FMTCodeFormatter`indentSize,
		3
	]

maxLineLengthPolicyAda[policies_Association]:=
	Lookup[policies,
		FMTCodeFormatter`maxLineLength,
		80
	]

preamblePolicyAda[policies_Association]:=
	Lookup[policies,
		FMTSymbols`fileHeaderComment,
		Function[{pkgName,trPkgName,fileName},
			(* by default, the preamble is empty,
				below is a simple template commented out *)
			{
			(*
				"--\n"<>
				"-- File "<>fileName<>"\n"<>
				"-- generated for package "<>trPkgName<>",\n"<>
				"-- from model package "<>pkgName<>"\n"<>
				"--",
				FMTCodeFormatter`lineBreakT[],
				FMTCodeFormatter`verticalSpaceT[]
			*)
			}
		]
	]


fileNamesAda[pkgName_String,policies_Association]:=
	fileNamesFromPkgNamePolicyAda[policies][pkgName]


withClausesAda[sys_,deps_List,policies_Association]:=Map[
	Function[dep,
		With[
			{
				sdep=nameFromSymbolPolicyAda[policies][dep,""]
			},
			
			{
				"with "<>sdep<>";",
				FMTCodeFormatter`lineBreakT[]
			}
		]
	],
	deps
]


numberDefinitionAda[name_String,policies_Association]:={
	"subtype "<>name,
	FMTCodeFormatter`canTrimT[" "],
	FMTCodeFormatter`noLineBreakT["is"],
	FMTCodeFormatter`canTrimT[" "],
	FMTCodeFormatter`noLineBreakT[nameFromSymbolPolicyAda[policies][FMTSymbols`integerT,""]],
	FMTCodeFormatter`noLineBreakT[";"],
	FMTCodeFormatter`lineBreakT[]
}


enumDefinitionAda[name_String,symbols_List,policies_Association]:=Module[
	{
		count,
		len=Length[symbols],
		result
	},
	
	result={
		"type "<>name,
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT["is"],
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`miniIndentIfContinuedT["("]
	};
	count=1;
	Scan[
		Function[e,
			AppendTo[result,
				FMTCodeFormatter`noLineBreakT[nameFromSymbolPolicyAda[policies][e,""]]
			];
			
			If[count<len,
				AppendTo[result,FMTCodeFormatter`noLineBreakT[","]];
				AppendTo[result,FMTCodeFormatter`canTrimT[" "]]
			];
			
			++count
		],
		symbols
	];
	AppendTo[result,FMTCodeFormatter`noLineBreakT[");"]];
	AppendTo[result,FMTCodeFormatter`lineBreakT[]];
	
	result
]


arrayDefinitionAda[name_String,def_,pkgName_String,policies_Association]:=With[
	{
		elementType=def[[1]],
		arraySize=def[[2]],
		indexOffset=policies[FMTOptions`arrayIndexOffset]
	},
	
	{
		"type "<>name,
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT["is"],
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT["array ("<>ToString[indexOffset]<>" .. "<>ToString[arraySize-1+indexOffset]<>")"],
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT["of"],
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT[nameFromSymbolPolicyAda[policies][elementType,pkgName]],
		FMTCodeFormatter`noLineBreakT[";"],
		FMTCodeFormatter`lineBreakT[]
	}
]


structureDefinitionAda[name_String,fields_List,pkgName_String,policies_Association]:=Module[
	{
		len=Length[fields],
		result
	},
	
	result={
		"type "<>name,
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT["is"],
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT["record"],
		FMTCodeFormatter`lineBreakT[],
		FMTCodeFormatter`moreIndentT[]
	};
	
	Scan[
		Function[f, (* f is name\[Rule]type *)
			With[
				{
					fieldName=f[[1]],
					fieldType=f[[2]]
				},
				
				AppendTo[result,
					{
						nameFromSymbolPolicyAda[policies][fieldName,pkgName],
						FMTCodeFormatter`canTrimT[" "],
						FMTCodeFormatter`noLineBreakT[":"],
						FMTCodeFormatter`canTrimT[" "],
						FMTCodeFormatter`noLineBreakT[nameFromSymbolPolicyAda[policies][fieldType,pkgName]],
						FMTCodeFormatter`noLineBreakT[";"],
						FMTCodeFormatter`lineBreakT[]
					}
				]
			]
		],
		fields
	];
	
	AppendTo[result,FMTCodeFormatter`lessIndentT[]];
	AppendTo[result,FMTCodeFormatter`noLineBreakT["end record;"]];
	AppendTo[result,FMTCodeFormatter`lineBreakT[]];
	
	result
]


typeDefinitionsAda[sys_,pkgName_String,policies_Association]:=Module[
	{
		defs=FMTSystem`packageTypeDefs[sys,pkgName],
		result={}
	},
	
	If[Length[defs]!=0,
		result={
			"-- Type definitions.",
			FMTCodeFormatter`verticalSpaceT[]
		}
	];
	
	Scan[
		Function[defRule,
			With[
				{
					sTypeName=nameFromSymbolPolicyAda[policies][defRule[[1]],pkgName],
					def=defRule[[2]]
				},
				
				appendGroupCommentAda[result,sys,defRule[[1]]];
				
				AppendTo[result,
					Switch[Head[def],
						Association, (* this is a number definition *)
							numberDefinitionAda[sTypeName,policies],
						
						FMTSymbols`enum,
							enumDefinitionAda[sTypeName,List@@def,policies],
						
						FMTSymbols`array,
							arrayDefinitionAda[sTypeName,def,pkgName,policies],
						
						FMTSymbols`structure,
							structureDefinitionAda[sTypeName,List@@def,pkgName,policies]
					]
				]
			]
		],
		Normal[defs]
	];
	
	AppendTo[result,FMTCodeFormatter`verticalSpaceT[]];
	
	result
]


objectSpecsAda[sys_,pkgName_String,policies_Association]:=Module[
	{
		defs=FMTSystem`packageDataDefs[sys,pkgName],
		result={}
	},
	
	If[Length[defs]!=0,
		result={
			"-- Data definitions.",
			FMTCodeFormatter`verticalSpaceT[]
		}
	];
	
	Scan[
		Function[defRule,
			With[
				{
					sObjName=nameFromSymbolPolicyAda[policies][defRule[[1]],pkgName],
					sTypeName=nameFromSymbolPolicyAda[policies][defRule[[2]],pkgName]
				},
				
				appendGroupCommentAda[result,sys,defRule[[1]]];
				
				AppendTo[result,
					{
						sObjName,
						FMTCodeFormatter`canTrimT[" "],
						FMTCodeFormatter`noLineBreakT[":"],
						FMTCodeFormatter`canTrimT[" "],
						FMTCodeFormatter`noLineBreakT[sTypeName],
						FMTCodeFormatter`noLineBreakT[";"],
						FMTCodeFormatter`lineBreakT[]
					}
				]
			]
		],
		Normal[defs]
	];
	
	If[result!={},
		AppendTo[result,FMTCodeFormatter`verticalSpaceT[]],
		result=Null
	];
	
	result
]


(* Helper, generates operation's profile only *)
operationProfileAda[pkgName_String,sOpName_String,profile_List,policies_Association]:=Module[
	{
		profileLen,
		count,
		result={}
	},
	
	AppendTo[result,"procedure "<>sOpName];
	profileLen=Length[profile];
	If[profileLen!=0,
		AppendTo[result,
			{
				FMTCodeFormatter`canTrimT[" "],
				FMTCodeFormatter`miniIndentIfContinuedT["("]
			}
		];
		
		count=1;
		Scan[
			Function[paramDef, (* paramDef is mode[name\[Rule]type] *)
				With[
					{
						mode=Head[paramDef],
						sParamName=nameFromSymbolPolicyAda[policies][paramDef[[1,1]],""],
						sTypeName=nameFromSymbolPolicyAda[policies][paramDef[[1,2]],pkgName]
					},
					
					AppendTo[result,
						{
							sParamName,
							FMTCodeFormatter`canTrimT[" "],
							":",
							FMTCodeFormatter`canTrimT[" "],
							Switch[mode,
								FMTSymbols`in,
									FMTCodeFormatter`noLineBreakT["in "<>sTypeName],
								FMTSymbols`inOut,
									FMTCodeFormatter`noLineBreakT["in out "<>sTypeName],
								FMTSymbols`out,
									FMTCodeFormatter`noLineBreakT["out "<>sTypeName]
							],
							If[count<profileLen,
								{
									FMTCodeFormatter`noLineBreakT[";"],
									FMTCodeFormatter`canTrimT[" "]
								},
								Nothing
							]
						}
					];
					++count
				]
			],
			profile
		];
		
		AppendTo[result,FMTCodeFormatter`noLineBreakT[")"]]
	];
	
	result
]


operationSpecsAda[sys_,pkgName_String,policies_Association]:=Module[
	{
		decls=FMTSystem`packageOperationDecls[sys,pkgName],
		result={}
	},
	
	If[Length[decls]!=0,
		result={
			"-- Operation declarations.",
			FMTCodeFormatter`verticalSpaceT[]
		}
	];
	
	Scan[
		Function[declRule,
			With[
				{
					sOpName=nameFromSymbolPolicyAda[policies][declRule[[1]],pkgName],
					profile=List@@declRule[[2]]
				},
				
				appendGroupCommentAda[result,sys,declRule[[1]]];
				
				AppendTo[result,operationProfileAda[pkgName,sOpName,profile,policies]];
				
				AppendTo[result,
					{
						FMTCodeFormatter`noLineBreakT[";"],
						FMTCodeFormatter`lineBreakT[]
					}
				]
			]
		],
		Normal[decls]
	];
	
	AppendTo[result,FMTCodeFormatter`verticalSpaceT[]];
	
	result
]


(* Helper, generates code for a single object name
	or some part of it, if it is an array access or structure field access
*)
objectCodeAda[obj_,pkgName_String,policies_Association]:=With[
	{
		wrapped=obj[[1]]
	},
	
	(* obj is objectT or objectLhsT or objectParentT,
		but it can wrap either a simple name (a symbol),
		or an array access or a structure field *)
	Switch[Head[wrapped],
		Symbol,
			FMTCodeFormatter`noLineBreakT[nameFromSymbolPolicyAda[policies][wrapped,pkgName]],
		
		FMTSymbols`partT,
			partCodeAda[wrapped,pkgName,policies],
		
		FMTSymbols`dotT,
			dotCodeAda[wrapped,pkgName,policies]
	]
]

(* Helper, generates code for array access names *)
partCodeAda[part_,pkgName_String,policies_Association]:=With[
	{
		parent=part[[1]],
		idx=part[[2]]
	},
	
	{
		objectCodeAda[parent,pkgName,policies],
		FMTCodeFormatter`noLineBreakT["("],
		expressionCodeAda[idx,Null,1,pkgName,policies],
		FMTCodeFormatter`noLineBreakT[")"]
	}
]

(* Helper, generates code for struct field names *)
dotCodeAda[dot_,pkgName_String,policies_Association]:=With[
	{
		parent=dot[[1]],
		field=dot[[2]]
	},
	
	{
		objectCodeAda[parent,pkgName,policies],
		FMTCodeFormatter`noLineBreakT["."],
		FMTCodeFormatter`noLineBreakT[nameFromSymbolPolicyAda[policies][field,""]]
	}
]


(* Helper, generates code for a single literal *)
literalCodeAda[lit_,outerExpr_Symbol,idx_,pkgName_String,policies_Association]:=Module[
	{
		value=lit[[1]],
		type=lit[[2]],
		result
	},
	
	Switch[type,
		FMTSymbols`integerT,
			If[value>=0,
				result=FMTCodeFormatter`noLineBreakT[ToString[value]],
				
				(* if value is negative and the containing expression is plusT,
					then instead of adding negative value, subtract the absolute value
					(plusCode already omitted the generation of plus symbol) *)
				Switch[outerExpr,
					FMTSymbols`plusT, (* for example x - 5, seen as plus[x,-5] *)
						If[idx==1,
							(* first value in addition is just a negative value, translate as is *)
							result=FMTCodeFormatter`noLineBreakT[ToString[value]],
							
							(* further values are subtractions instead of negations *)
							result={
								FMTCodeFormatter`noLineBreakT["-"],
								FMTCodeFormatter`canTrimT[" "],
								FMTCodeFormatter`noLineBreakT[ToString[-value]]
							}
						],
					
					FMTSymbols`timesT|FMTSymbols`divT, (* for example x * -5 or x / -5 *)
						If[idx==1,
							(* first value in multiplication or division is just a negative value,
								translate as is *)
							result=FMTCodeFormatter`noLineBreakT[ToString[value]],
							
							(* further values need parens *)
							result={
								FMTCodeFormatter`noLineBreakT["("],
								FMTCodeFormatter`noLineBreakT[ToString[value]],
								FMTCodeFormatter`noLineBreakT[")"]
							}
						],
					
					FMTSymbols`negT, (* for example x - -5 *)
						result={
							FMTCodeFormatter`noLineBreakT["("],
							FMTCodeFormatter`noLineBreakT[ToString[value]],
							FMTCodeFormatter`noLineBreakT[")"]
						},
					
					_,
						(* otherwise, translate the negative value as is *)
						result=FMTCodeFormatter`noLineBreakT[ToString[value]]
				]
			],
		
		FMTSymbols`booleanT,
			result=FMTCodeFormatter`noLineBreakT[If[value,"True","False"]],
		
		_,
			(* otherwise, it is an enum literal *)
			result=FMTCodeFormatter`noLineBreakT[nameFromSymbolPolicyAda[policies][value,pkgName]]
	];
	
	result
]


(* Helper, generates code for a negated value (which is object or literal) *)
negCodeAda[neg_,outerExpr_Symbol,idx_Integer,pkgName_String,policies_Association]:=Module[
	{
		value=neg[[1]],
		result
	},
	
	Switch[outerExpr,
		FMTSymbols`plusT, (* for example 5 - x, seen as plus[5,neg[x]] *)
			(* if the negated value is part of the plusT operation,
				it needs space between minus sign and the actual value,
				as the minus sign is generated as an operator (instead of plus),
				not a negator
				but only for second and later elements of the plus operation
			*)
			result={
				FMTCodeFormatter`noLineBreakT["-"],
				If[idx>1,
					FMTCodeFormatter`canTrimT[" "],
					Nothing
				],
				expressionCodeAda[value,FMTSymbols`negT,1,pkgName,policies]
			},
		
		FMTSymbols`negT, (* multiple negation is allowed, but needs parens *)
			result={
				FMTCodeFormatter`noLineBreakT["(-"],
				expressionCodeAda[value,FMTSymbols`negT,1,pkgName,policies],
				FMTCodeFormatter`noLineBreakT[")"]
			},
		
		_, (* regular negation *)
			result={
				FMTCodeFormatter`noLineBreakT["-"],
				expressionCodeAda[value,FMTSymbols`negT,1,pkgName,policies]
			}
	];
	
	result
]


(* Helper, generates code for a plus operator *)
plusCodeAda[plus_,pkgName_String,policies_Association]:=Module[
	{
		len=Length[plus],
		count,
		result={}
	},
	
	count=1;
	Scan[
		Function[part, (* part is a nested expression *)
			(* recursively translate all involved parts *)
			AppendTo[result,expressionCodeAda[part,FMTSymbols`plusT,count,pkgName,policies]];
			
			If[count<len,
				If[Head[plus[[count+1]]]===FMTSymbols`negT||
					(Head[plus[[count+1]]]===FMTSymbols`literalT&&
						plus[[count+1,2]]===FMTSymbols`integerT&&
						plus[[count+1,1]]<0),
					
					(* the next element is either negated or it is a negative integer literal
						\[Rule] do not generate any operator sign, negCode and literalCode will
						generate the minus sign instead *)
					AppendTo[result,FMTCodeFormatter`canTrimT[" "]],
					
					(* use plus sign for regular additions *)
					AppendTo[result,
						{
							FMTCodeFormatter`canTrimT[" "],
							FMTCodeFormatter`noLineBreakT["+"],
							FMTCodeFormatter`canTrimT[" "]
						}
					]
				]
			];
			
			++count
		],
		plus
	];
	
	result
]


(* Helper, generates code for a times operator *)
timesCodeAda[times_,pkgName_String,policies_Association]:=Module[
	{
		len=Length[times],
		count,
		result={}
	},
	
	count=1;
	Scan[
		Function[part, (* part is a nested expression *)
			(* recursively translate all involved parts *)
			AppendTo[result,expressionCodeAda[part,FMTSymbols`timesT,count,pkgName,policies]];
			
			If[count<len,
				AppendTo[result,
					{
						FMTCodeFormatter`canTrimT[" "],
						FMTCodeFormatter`noLineBreakT["*"],
						FMTCodeFormatter`canTrimT[" "]
					}
				]
			];
			
			++count
		],
		times
	];
	
	result
]


(* Helper, generates code for a div operator *)
divCodeAda[div_,pkgName_String,policies_Association]:=Module[
	{
		len=Length[div],
		count,
		result={}
	},
	
	count=1;
	Scan[
		Function[part, (* part is a nested expression *)
			(* recursively translate all involved parts *)
			AppendTo[result,expressionCodeAda[part,FMTSymbols`divT,count,pkgName,policies]];
			
			If[count<len,
				AppendTo[result,
					{
						FMTCodeFormatter`canTrimT[" "],
						FMTCodeFormatter`noLineBreakT["/"],
						FMTCodeFormatter`canTrimT[" "]
					}
				]
			];
			
			++count
		],
		div
	];
	
	result
]


(* Helper, generates code for a generic operator: and, or, \[Equal], \[NotEqual], <, \[LessEqual], >, \[GreaterEqual] *)
generalOperatorCodeAda[op_,pkgName_String,policies_Association]:=Module[
	{
		len=Length[op],
		count,
		result={}
	},
	
	count=1;
	Scan[
		Function[part, (* part is a nested expression *)
			(* recursively translate all involved parts *)
			AppendTo[result,expressionCodeAda[part,Head[op],count,pkgName,policies]];
			
			If[count<len,
				AppendTo[result,
					{
						FMTCodeFormatter`canTrimT[" "],
						Switch[Head[op],
							FMTSymbols`andT,
								FMTCodeFormatter`noLineBreakT["and"],
							
							FMTSymbols`orT,
								FMTCodeFormatter`noLineBreakT["or"],
							
							FMTSymbols`equalT,
								FMTCodeFormatter`noLineBreakT["="],
							
							FMTSymbols`unequalT,
								FMTCodeFormatter`noLineBreakT["/="],
							
							FMTSymbols`lessT,
								FMTCodeFormatter`noLineBreakT["<"],
							
							FMTSymbols`lessEqualT,
								FMTCodeFormatter`noLineBreakT["<="],
							
							FMTSymbols`greaterT,
								FMTCodeFormatter`noLineBreakT[">"],
							
							FMTSymbols`greaterEqualT,
								FMTCodeFormatter`noLineBreakT[">="]
						],
						FMTCodeFormatter`canTrimT[" "]
					}
				]
			];
			
			++count
		],
		op
	];
	
	result
]


(* Helper, generates code for a not operator *)
notCodeAda[not_,pkgName_String,policies_Association]:={
	FMTCodeFormatter`noLineBreakT["not "], (* no line breaks here *)
	expressionCodeAda[not[[1]],FMTSymbols`notT,1,pkgName,policies]
}


(* Helper, generates code for the whole expression,
	outerExpr is a Head of the outer (containing) expression (or Null),
	idx is an index within the outer expression,
	both are used to decide whether negation sign needs space or whether parens are needed *)
expressionCodeAda[expr_,outerExpr_Symbol,idx_Integer,pkgName_String,policies_Association]:=Module[
	{
		result
	},
	
	Switch[Head[expr],
		FMTSymbols`objectT,
			result=objectCodeAda[expr,pkgName,policies],
		
		FMTSymbols`literalT,
			result=literalCodeAda[expr,outerExpr,idx,pkgName,policies],
		
		FMTSymbols`negT,
			result=negCodeAda[expr,outerExpr,idx,pkgName,policies],
		
		FMTSymbols`plusT,
			If[MemberQ[{FMTSymbols`timesT,FMTSymbols`divT,FMTSymbols`negT},outerExpr],
				(* parens are needed for the whole plus operation *)
				result={
					FMTCodeFormatter`noLineBreakT["("],
					plusCodeAda[expr,pkgName,policies],
					FMTCodeFormatter`noLineBreakT[")"]
				},
				
				(* no parens needed *)
				result=plusCodeAda[expr,pkgName,policies]
			],
		
		FMTSymbols`timesT,
			If[outerExpr===FMTSymbols`negT,
				(* parens are needed for the whole times operation *)
				result={
					FMTCodeFormatter`noLineBreakT["("],
					timesCodeAda[expr,pkgName,policies],
					FMTCodeFormatter`noLineBreakT[")"]
				},
				
				(* no parens needed *)
				result=timesCodeAda[expr,pkgName,policies]
			],
		
		FMTSymbols`divT,
			If[outerExpr===FMTSymbols`negT,
				(* parens are needed for the whole div operation *)
				result={
					FMTCodeFormatter`noLineBreakT["("],
					divCodeAda[expr,pkgName,policies],
					FMTCodeFormatter`noLineBreakT[")"]
				},
				
				(* no parens needed *)
				result=divCodeAda[expr,pkgName,policies]
			],
		
		FMTSymbols`andT|FMTSymbols`orT|
		FMTSymbols`equalT|FMTSymbols`unequalT|
		FMTSymbols`lessT|FMTSymbols`lessEqualT|
		FMTSymbols`greaterT|FMTSymbols`greaterEqualT,
			If[outerExpr=!=Null,
				(* just in case, parens are needed for the whole and operation *)
				result={
					FMTCodeFormatter`noLineBreakT["("],
					generalOperatorCodeAda[expr,pkgName,policies],
					FMTCodeFormatter`noLineBreakT[")"]
				},
				
				(* no parens needed *)
				result=generalOperatorCodeAda[expr,pkgName,policies]
			],
		
		FMTSymbols`notT,
			result=notCodeAda[expr,pkgName,policies]
	];
	
	result
]


(* Helper statement-related code generators, separate functions for each statement type *)

userCommentAda[comment_FMTSymbols`commentT]:=With[
	{
		lines=List@@comment
	},
	
	Map[
		Function[line,
			{
				"-- "<>line,
				FMTCodeFormatter`lineBreakT[]
			}
		],
		lines
	]
]

appendGroupCommentAda[result_,sys_,entityName_]:=With[
	{
		comment=FMTSystem`fmtGetComment[sys,entityName]
	},
	
	If[Head[comment]===FMTSymbols`commentT,
		AppendTo[result,FMTCodeFormatter`verticalSpaceT[]];
		AppendTo[result,userCommentAda[comment]],
		
		If[comment===Null,
			(* this means that there is no comment,
				but the entity is the first in its group *)
			AppendTo[result,FMTCodeFormatter`verticalSpaceT[]]
		]
	]
]

SetAttributes[appendGroupCommentAda,HoldFirst]

emptyStatementCodeAda[]:={
	"null;",
	FMTCodeFormatter`lineBreakT[]
}

compoundStatementCodeAda[comp_FMTSymbols`compoundT,pkgName_String,policies_Association]:=Module[
	{
		len=Length[comp],
		count,
		addVerticalSpaceBefore,
		addVerticalSpaceAfter,
		result={}
	},
	
	count=1;
	Scan[
		Function[st, (* st is a single statement within a compound *)
			(* add vertical space around control structures and procedure calls *)
			If[MemberQ[{FMTSymbols`ifT,FMTSymbols`whileT,FMTSymbols`switchT,FMTSymbols`callT},
					Head[st]],
					
				addVerticalSpaceBefore=1<count;
				addVerticalSpaceAfter=count<len
			];
			++count;
			
			If[Head[st]=!=FMTSymbols`statementT, (* avoid empty statements in compounds *)
				If[addVerticalSpaceBefore,
					AppendTo[result,FMTCodeFormatter`verticalSpaceT[]]
				];
				
				AppendTo[result,bodyCodeFromStatementAda[st,pkgName,policies]];
				
				If[addVerticalSpaceAfter,
					AppendTo[result,FMTCodeFormatter`verticalSpaceT[]]
				]
			]
		],
		comp
	];
	
	result
]

assignmentCodeAda[st_FMTSymbols`assignmentT,pkgName_String,policies_Association]:=With[
	{
		lhsObject=st[[1]],
		rhsExpression=st[[2]]
	},
	
	{
		objectCodeAda[lhsObject,pkgName,policies],
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT[":="],
		FMTCodeFormatter`canTrimT[" "],
		expressionCodeAda[rhsExpression,Null,1,pkgName,policies],
		FMTCodeFormatter`noLineBreakT[";"],
		FMTCodeFormatter`lineBreakT[]
	}
]

ifCodeAda[st_FMTSymbols`ifT,pkgName_String,policies_Association]:=With[
	{
		cond=st[[1]],
		thenBranch=st[[2]],
		elseBranch=If[Length[st]>2,st[[3]],Null]
	},
	
	{
		FMTCodeFormatter`noLineBreakT["if "],
		expressionCodeAda[cond,Null,1,pkgName,policies],
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT["then"],
		FMTCodeFormatter`lineBreakT[],
		FMTCodeFormatter`moreIndentT[],
		bodyCodeFromStatementAda[thenBranch,pkgName,policies],
		FMTCodeFormatter`lessIndentT[],
		If[elseBranch=!=Null,
			{
				FMTCodeFormatter`noLineBreakT["else"],
				FMTCodeFormatter`lineBreakT[],
				FMTCodeFormatter`moreIndentT[],
				bodyCodeFromStatementAda[elseBranch,pkgName,policies],
				FMTCodeFormatter`lessIndentT[]
			},
			Nothing
		],
		FMTCodeFormatter`noLineBreakT["end if;"],
		FMTCodeFormatter`lineBreakT[]
	}
]

whileCodeAda[while_FMTSymbols`whileT,pkgName_String,policies_Association]:=With[
	{
		cond=while[[1]],
		loopBranch=while[[2]]
	},
	
	{
		FMTCodeFormatter`noLineBreakT["while "],
		expressionCodeAda[cond,Null,1,pkgName,policies],
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT["loop"],
		FMTCodeFormatter`lineBreakT[],
		FMTCodeFormatter`moreIndentT[],
		bodyCodeFromStatementAda[loopBranch,pkgName,policies],
		FMTCodeFormatter`lessIndentT[],
		FMTCodeFormatter`noLineBreakT["end loop;"],
		FMTCodeFormatter`lineBreakT[]
	}
]

whenValuesAda[values_,pkgName_String,policies_Association]:=
	If[values===FMTSymbols`defaultT[],
		FMTCodeFormatter`noLineBreakT["others"],
		
		If[Head[values]===FMTSymbols`literalT,
			(* it is a single value *)
			nameFromSymbolPolicyAda[policies][values[[1]],pkgName],
			
			(* it is a list of alternatives *)
			(* Note, Riffle cannot treat its second List argument as a single entity,
				so here an explicit Scan is used to insert " | " between values *)
			Module[
				{
					len=Length[values],
					count,
					result={}
				},
				
				count=1;
				Scan[
					Function[v, (* v is a single literalT[enumName,enumType] from the list of alternatives *)
						AppendTo[result,nameFromSymbolPolicyAda[policies][v[[1]],pkgName]];
						If[count++<len,
							AppendTo[result,
								{
									FMTCodeFormatter`canTrimT[" "],
									FMTCodeFormatter`noLineBreakT["|"],
									FMTCodeFormatter`canTrimT[" "]
								}
							]
						]
					],
					List@@values
				];
				
				result
			]
		]
	]

switchCodeAda[switch_FMTSymbols`switchT,pkgName_String,policies_Association]:=Module[
	{
		expr=switch[[1]],
		result
	},
	
	result={
		FMTCodeFormatter`noLineBreakT["case "],
		expressionCodeAda[expr,Null,1,pkgName,policies],
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT["is"],
		FMTCodeFormatter`lineBreakT[],
		FMTCodeFormatter`moreIndentT[]
	};
	
	AppendTo[result,
		Riffle[
			Map[
				Function[vb, (* vb is {values,branch} *)
					With[
						{
							values=vb[[1]],
							branch=vb[[2]]
						},
						
						{
							FMTCodeFormatter`noLineBreakT["when "],
							whenValuesAda[values,pkgName,policies],
							FMTCodeFormatter`canTrimT[" "],
							FMTCodeFormatter`noLineBreakT["=>"],
							FMTCodeFormatter`lineBreakT[],
							FMTCodeFormatter`moreIndentT[],
							bodyCodeFromStatementAda[branch,pkgName,policies],
							FMTCodeFormatter`lessIndentT[]
						}
					]
				],
				Partition[List@@switch[[2;;]],2]
			],
			FMTCodeFormatter`verticalSpaceT[]
		]
	];
	
	AppendTo[result,
		{
			FMTCodeFormatter`lessIndentT[],
			FMTCodeFormatter`noLineBreakT["end case;"],
			FMTCodeFormatter`lineBreakT[]
		}
	];
	
	result
]

callCodeAda[call_FMTSymbols`callT,pkgName_String,policies_Association]:=Module[
	{
		qProcName=call[[1]],
		formals=call[[2]],
		actuals=call[[3]],
		len,
		count,
		result
	},
	
	len=Length[actuals];
	
	result={nameFromSymbolPolicyAda[policies][qProcName,pkgName]};
	
	If[len>0,
		AppendTo[result,
			{
				FMTCodeFormatter`canTrimT[" "],
				FMTCodeFormatter`miniIndentIfContinuedT["("]
			}
		];
		
		count=1;
		Scan[
			Function[actual,
				AppendTo[result,expressionCodeAda[actual,Null,1,pkgName,policies]];
				
				If[count<len,
					AppendTo[result,
						{
							FMTCodeFormatter`noLineBreakT[","],
							FMTCodeFormatter`canTrimT[" "]
						}
					]
				];
				
				++count
			],
			actuals
		];
		
		AppendTo[result,FMTCodeFormatter`noLineBreakT[")"]]
	];
	
	AppendTo[result,
		{
			FMTCodeFormatter`noLineBreakT[";"],
			FMTCodeFormatter`lineBreakT[]
		}
	];
	
	result
]

bodyCodeFromStatementAda[st_,pkgName_String,policies_Association]:=Module[
	{
		result
	},
	
	Switch[Head[st],
		FMTSymbols`commentT,
			(* this is a user-injected comment *)
			result=userCommentAda[st],
		
		FMTSymbols`statementT,
			(* this is a Null statement (perhaps a null procedure) *)
			result=emptyStatementCodeAda[],
		
		FMTSymbols`compoundT,
			(* this is a compound statement (a sequence of statements) *)
			result=compoundStatementCodeAda[st,pkgName,policies],
		
		FMTSymbols`assignmentT,
			result=assignmentCodeAda[st,pkgName,policies],
		
		FMTSymbols`ifT,
			result=ifCodeAda[st,pkgName,policies],
		
		FMTSymbols`whileT,
			result=whileCodeAda[st,pkgName,policies],
		
		FMTSymbols`switchT,
			result=switchCodeAda[st,pkgName,policies],
		
		FMTSymbols`callT,
			result=callCodeAda[st,pkgName,policies]
	];
	
	result
]


operationDefsAda[sys_,pkgName_String,policies_Association]:=Module[
	{
		decls=FMTSystem`packageOperationDecls[sys,pkgName],
		result={}
	},
	
	Scan[
		Function[declRule,
			With[
				{
					qOpName=declRule[[1]],
					sOpName=nameFromSymbolPolicyAda[policies][declRule[[1]],pkgName],
					profile=List@@declRule[[2]],
					localVars=FMTSystem`operationLocalVars[sys,declRule[[1]]],
					opBody=FMTSystem`resolvedOperationBody[sys,pkgName,declRule[[1]],False,{}]
				},
				
				If[opBody=!=Null,
					AppendTo[result,
						{
							operationProfileAda[pkgName,sOpName,profile,policies],
							FMTCodeFormatter`canTrimT[" "],
							FMTCodeFormatter`noLineBreakT["is"],
							FMTCodeFormatter`lineBreakT[]
						}
					];
					
					(* define local variables *)
					If[localVars!={},
						AppendTo[result,FMTCodeFormatter`moreIndentT[]];
						Scan[
							Function[varDef, (* varDef is name\[Rule]type *)
								With[
									{
										name=nameFromSymbolPolicyAda[policies][varDef[[1]],pkgName],
										type=nameFromSymbolPolicyAda[policies][varDef[[2]],pkgName]
									},
									
									AppendTo[result,
										{
											name,
											FMTCodeFormatter`canTrimT[" "],
											":",
											FMTCodeFormatter`canTrimT[" "],
											FMTCodeFormatter`noLineBreakT[type],
											FMTCodeFormatter`noLineBreakT[";"],
											FMTCodeFormatter`lineBreakT[]
										}
									]
								]
							],
							localVars
						];
						AppendTo[result,FMTCodeFormatter`lessIndentT[]];
					];
					
					(* generate procedure's body *)
					AppendTo[result,
						{
							"begin",
							FMTCodeFormatter`lineBreakT[],
							FMTCodeFormatter`moreIndentT[]
						}
					];
					
					AppendTo[result,bodyCodeFromStatementAda[opBody,pkgName,policies]];
					
					AppendTo[result,
						{
							FMTCodeFormatter`lessIndentT[],
							"end "<>sOpName<>";",
							FMTCodeFormatter`verticalSpaceT[]
						}
					]
				]
			]
		],
		Normal[decls]
	];
	
	If[result!={},
		{
			"-- Operation definitions.",
			FMTCodeFormatter`verticalSpaceT[],
			
			result
		},
		
		(* return Null if there are no procedure bodies generated in this package *)
		Null
	]
]


generateSpecAda[sys_,pkgName_String,fileName_String,specDeps_List,hasBody_,policies_Association]:=Module[
	{
		sPkgName=nameFromSymbolPolicyAda[policies][pkgName,""],
		indentSize=indentSizePolicyAda[policies],
		maxLineLength=maxLineLengthPolicyAda[policies],
		preamble=preamblePolicyAda[policies],
		pkgComment,
		objectDeclarations,
		noBodyInfo
	},
	
	pkgComment=FMTSystem`fmtGetComment[sys,pkgName];
	objectDeclarations=objectSpecsAda[sys,pkgName,policies];
	
	If[Not[hasBody],
		noBodyInfo={
			"-- This package does not need body.",
			FMTCodeFormatter`lineBreakT[]
		};
		
		If[objectDeclarations===Null,
			AppendTo[noBodyInfo,"pragma Pure ("<>sPkgName<>");"];
			AppendTo[noBodyInfo,FMTCodeFormatter`lineBreakT[]]
		];
		
		AppendTo[noBodyInfo,FMTCodeFormatter`verticalSpaceT[]],
		
		noBodyInfo=Nothing
	];
	
	FMTCodeFormatter`format[Flatten@{
		(* generate file preamble, if requested by policies (by default empty) *)
		preamble[pkgName,sPkgName,fileName],
		
		(* generate with clauses for dependent packages *)
		withClausesAda[sys,specDeps,policies],
		
		FMTCodeFormatter`verticalSpaceT[],
		If[Head[pkgComment]===FMTSymbols`commentT,
			userCommentAda[pkgComment],
			Nothing
		],
		"package "<>sPkgName,
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT["is"],
		FMTCodeFormatter`lineBreakT[],
		FMTCodeFormatter`moreIndentT[],
		FMTCodeFormatter`verticalSpaceT[],
		
		noBodyInfo,
		
		(* generate type definitions *)
		typeDefinitionsAda[sys,pkgName,policies],
		
		(* generate object definitions *)
		(* Note: object definitions are generated before operations,
			because operation profiles can refer to objects in their pre- and post-conditions.
			This means that objects (at least those that are referred) need to be defined
			first in the model and this convention is preserved in the generated code.
		*)
		objectDeclarations,
		
		(* generate operation prototypes *)
		operationSpecsAda[sys,pkgName,policies],
		
		FMTCodeFormatter`verticalSpaceT[],
		FMTCodeFormatter`lessIndentT[],
		"end ",FMTCodeFormatter`noLineBreakT[sPkgName],FMTCodeFormatter`noLineBreakT[";"],
		FMTCodeFormatter`lineBreakT[]
	},
	<|
		FMTCodeFormatter`indentSize->indentSize,
		FMTCodeFormatter`maxLineLength->maxLineLength
	|>]
]


generateBodyAda[sys_,pkgName_String,fileName_String,bodyDeps_List,policies_Association]:=Module[
	{
		sPkgName=nameFromSymbolPolicyAda[policies][pkgName,""],
		indentSize=indentSizePolicyAda[policies],
		maxLineLength=maxLineLengthPolicyAda[policies],
		preamble=preamblePolicyAda[policies],
		operations,
		result
	},
	
	(* generate operation definitions *)
	operations=operationDefsAda[sys,pkgName,policies];
	
	If[operations=!=Null,
		result=FMTCodeFormatter`format[Flatten@{
			(* generate file preamble, if requested by policies (by default empty) *)
			preamble[pkgName,sPkgName,fileName],
		
			(* generate with clauses for dependent packages *)
			withClausesAda[sys,bodyDeps,policies],
			
			FMTCodeFormatter`verticalSpaceT[],
			"package body "<>sPkgName,
			FMTCodeFormatter`canTrimT[" "],
			FMTCodeFormatter`noLineBreakT["is"],
			FMTCodeFormatter`lineBreakT[],
			FMTCodeFormatter`moreIndentT[],
			FMTCodeFormatter`verticalSpaceT[],
			
			operations,
			
			FMTCodeFormatter`verticalSpaceT[],
			FMTCodeFormatter`lessIndentT[],
			"end "<>sPkgName<>";",
			FMTCodeFormatter`lineBreakT[]
		},
		<|
			FMTCodeFormatter`indentSize->indentSize,
			FMTCodeFormatter`maxLineLength->maxLineLength
		|>],
		
		result=Null
	];
	
	result
]


fmtGenerateAda[sys_,pkgName_String,policies_Association]:=Module[
	{
		fNames,
		dependencies=FMTSystem`dependentPackages[sys,pkgName],
		spec,
		body,
		result=<||>
	},
	
	fNames=fileNamesAda[pkgName,policies];
	
	body=generateBodyAda[sys,pkgName,fNames[[2]],dependencies[[2]],policies];
	spec=generateSpecAda[sys,pkgName,fNames[[1]],dependencies[[1]],body=!=Null,policies];
	
	result[fNames[[1]]]=spec;
	result[fNames[[2]]]=body;
	
	result
]


End[];


EndPackage[];
