(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


Needs["FMTCodeFormatter`"];
Needs["FMTOptions`"];
Needs["FMTSymbols`"];
Needs["FMTSystem`"];


BeginPackage["FMTGeneratorCpp`"];


fileNamesCpp::usage="fileNamesCpp[pkgName,policies] Returns a pair of file names for spec and body for the given package and translation policies.";


fmtGenerateCpp::usage="fmtGenerateCpp[sys,pkgName,policies] Returns, <|fileNameSpec->codeSpec,fileNameBody->codeBody|> for the given package.\n"<>
"The policies is a possibly empty association with functions for various translation steps, these policies override their respective defaults.";


Begin["Private`"];


(* Helpers *)

normalizePackageSeparatorsCpp[n_String]:=StringRiffle[StringSplit[n,"`"],"::"]

(* default symbol translator, expects two parameters:
	- symbol or symbol name
	- name of current package (used to remove it from qualified part of the translated name)
*)
nameToStringCpp[n_,currentPackageName_String,policies_Association]:=Module[
	{
		str,
		previousWasLowerCase=False,
		temp={}
	},
	
	If[Head[n]===Symbol,
		Switch[n,
			FMTSymbols`integerT,
				str=integerPolicyCpp[policies],
			FMTSymbols`booleanT,
				str="bool",
			_,
				str=ToString[n]
		],
		
		str=n
	];
	
	str=StringTrim[str,currentPackageName];
	
	Scan[
		Function[c,
			If[previousWasLowerCase&&(UpperCaseQ[c]||DigitQ[c]),
				AppendTo[temp,"_"]
			];
			AppendTo[temp,ToLowerCase[c]];
			previousWasLowerCase=LowerCaseQ[c]
		],
		Characters[str]
	];
	
	normalizePackageSeparatorsCpp[StringJoin[temp]]
]

nameToFileRootCpp[n_String]:=ToLowerCase[StringReplace[n,"::"->"-"]]


defaultMemCpyCpp[dst_String,src_String,aType_String,eType_String,size_Integer]:=
	{
		FMTCodeFormatter`noLineBreakT["// array copy: "<>dst<>" = "<>src<>";"],
		FMTCodeFormatter`lineBreakT[],
		FMTCodeFormatter`noLineBreakT["std::memmove"],
		FMTCodeFormatter`noLineBreakT["((void *)"<>dst<>","],
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT["(void *)"<>src<>","],
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT["sizeof("<>aType<>"));"],
		FMTCodeFormatter`lineBreakT[]
	}


(* Policy helpers, they return default policy if there is no provided by the user. *)

fileNamesFromPkgNamePolicyCpp[policies_Association]:=
	Lookup[policies,
		FMTSymbols`fileNamesFromPkgName,
		Function[pkgName,
			With[
				{
					fileRoot=nameToFileRootCpp[nameFromSymbolPolicyCpp[policies][pkgName,""]]
				},
				
				{fileRoot<>".h",fileRoot<>".cpp"}
			]
		]
	]

nameFromSymbolPolicyCpp[policies_Association]:=With[
	{
		userPolicy=Lookup[policies,FMTSymbols`nameFromSymbol,Null]
	},
	
	Function[{n,pkgName},
		If[Head[userPolicy]===Association,
			(* userPolicy is a lookup table, try there first *)
			Lookup[userPolicy,n,nameToStringCpp[n,pkgName,policies]],
			
			If[userPolicy=!=Null,
				(* userPolicy is a function (hopefully) *)
				userPolicy[n,pkgName],
				
				(* fall back to default converter *)
				nameToStringCpp[n,pkgName,policies]
			]
		]
	]
]

indentSizePolicyCpp[policies_Association]:=
	Lookup[policies,
		FMTCodeFormatter`indentSize,
		4
	]

maxLineLengthPolicyCpp[policies_Association]:=
	Lookup[policies,
		FMTCodeFormatter`maxLineLength,
		80
	]

integerPolicyCpp[policies_Association]:=
	Lookup[policies,
		FMTSymbols`integerTargetType,
		"std::int32_t"
	]

memCpyPolicyCpp[policies_Association]:=
	Lookup[policies,
		FMTSymbols`arrayCopyOperation,
		defaultMemCpyCpp
	]

preamblePolicyCpp[policies_Association]:=
	Lookup[policies,
		FMTSymbols`fileHeaderComment,
		Function[{pkgName,trPkgName,fileName},
			(* by default, the preamble is empty,
				below is a simple template commented out *)
			{
			(*
				"//\n"<>
				"// File "<>fileName<>"\n"<>
				"// generated for package "<>trPkgName<>",\n"<>
				"// from model package "<>pkgName<>"\n"<>
				"//",
				FMTCodeFormatter`lineBreakT[],
				FMTCodeFormatter`verticalSpaceT[]
			*)
			}
		]
	]


fileNamesCpp[pkgName_String,policies_Association]:=
	fileNamesFromPkgNamePolicyCpp[policies][pkgName]


includesCpp[deps_List,policies_Association]:=Map[
	Function[dep,
		With[
			{
				sdep=nameFromSymbolPolicyCpp[policies][dep,""]
			},
			
			{
				"#include \""<>sdep<>".h\"",
				FMTCodeFormatter`lineBreakT[]
			}
		]
	],
	deps
]


numberDefinitionCpp[name_String,policies_Association]:={
	"typedef",
	FMTCodeFormatter`canTrimT[" "],
	nameFromSymbolPolicyCpp[policies][FMTSymbols`integerT,""],
	FMTCodeFormatter`canTrimT[" "],
	FMTCodeFormatter`noLineBreakT[name],
	FMTCodeFormatter`noLineBreakT[";"],
	FMTCodeFormatter`lineBreakT[]
}


enumDefinitionCpp[name_String,symbols_List,policies_Association]:=Module[
	{
		count,
		len=Length[symbols],
		result
	},
	
	result={
		"enum "<>name,
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`miniIndentIfContinuedT["{"]
	};
	count=1;
	Scan[
		Function[e,
			AppendTo[result,
				FMTCodeFormatter`noLineBreakT[nameFromSymbolPolicyCpp[policies][e,""]]
			];
			
			If[count<len,
				AppendTo[result,FMTCodeFormatter`noLineBreakT[","]];
				AppendTo[result,FMTCodeFormatter`canTrimT[" "]]
			];
			
			++count
		],
		symbols
	];
	AppendTo[result,FMTCodeFormatter`noLineBreakT["};"]];
	AppendTo[result,FMTCodeFormatter`lineBreakT[]];
	
	result
]


arrayDefinitionCpp[name_String,def_,pkgName_String,policies_Association]:=With[
	{
		elementType=def[[1]],
		arraySize=def[[2]]
	},
	
	{
		"typedef",
		FMTCodeFormatter`canTrimT[" "],
		nameFromSymbolPolicyCpp[policies][elementType,pkgName],
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT[name],
		FMTCodeFormatter`noLineBreakT["["<>ToString[arraySize]<>"]"],
		FMTCodeFormatter`noLineBreakT[";"],
		FMTCodeFormatter`lineBreakT[]
	}
]


structureDefinitionCpp[name_String,fields_List,pkgName_String,policies_Association]:=Module[
	{
		len=Length[fields],
		result
	},
	
	result={
		"struct "<>name,
		FMTCodeFormatter`lineBreakT[],
		FMTCodeFormatter`noLineBreakT["{"],
		FMTCodeFormatter`lineBreakT[],
		FMTCodeFormatter`moreIndentT[]
	};
	
	Scan[
		Function[f, (* f is name\[Rule]type *)
			With[
				{
					fieldName=f[[1]],
					fieldType=f[[2]]
				},
				
				AppendTo[result,
					{
						nameFromSymbolPolicyCpp[policies][fieldType,pkgName],
						FMTCodeFormatter`canTrimT[" "],
						FMTCodeFormatter`noLineBreakT[nameFromSymbolPolicyCpp[policies][fieldName,pkgName]],
						FMTCodeFormatter`noLineBreakT[";"],
						FMTCodeFormatter`lineBreakT[]
					}
				]
			]
		],
		fields
	];
	
	AppendTo[result,FMTCodeFormatter`lessIndentT[]];
	AppendTo[result,FMTCodeFormatter`noLineBreakT["};"]];
	AppendTo[result,FMTCodeFormatter`lineBreakT[]];
	
	result
]


typeDefinitionsCpp[sys_,pkgName_String,policies_Association]:=Module[
	{
		defs=FMTSystem`packageTypeDefs[sys,pkgName],
		result={}
	},
	
	If[Length[defs]!=0,
		result={
			"// Type definitions.",
			FMTCodeFormatter`verticalSpaceT[]
		}
	];
	
	Scan[
		Function[defRule,
			With[
				{
					sTypeName=nameFromSymbolPolicyCpp[policies][defRule[[1]],pkgName],
					def=defRule[[2]]
				},
				
				appendGroupCommentCpp[result,sys,defRule[[1]]];
				
				AppendTo[result,
					Switch[Head[def],
						Association, (* this is a number definition *)
							numberDefinitionCpp[sTypeName,policies],
						
						FMTSymbols`enum,
							enumDefinitionCpp[sTypeName,List@@def,policies],
						
						FMTSymbols`array,
							arrayDefinitionCpp[sTypeName,def,pkgName,policies],
						
						FMTSymbols`structure,
							structureDefinitionCpp[sTypeName,List@@def,pkgName,policies]
					]
				]
			]
		],
		Normal[defs]
	];
	
	AppendTo[result,FMTCodeFormatter`verticalSpaceT[]];
	
	result
]


objectSpecsCpp[sys_,inHeader_Symbol,pkgName_String,policies_Association]:=Module[
	{
		defs=FMTSystem`packageDataDefs[sys,pkgName],
		result={}
	},
	
	If[Length[defs]!=0,
		If[inHeader,
			result={
				"// Data declarations.",
				FMTCodeFormatter`verticalSpaceT[]
			},
			
			result={
				"// Data definitions.",
				FMTCodeFormatter`verticalSpaceT[]
			}
		]
	];
	
	Scan[
		Function[defRule,
			With[
				{
					sObjName=nameFromSymbolPolicyCpp[policies][defRule[[1]],pkgName],
					sTypeName=nameFromSymbolPolicyCpp[policies][defRule[[2]],pkgName]
				},
				
				If[inHeader,
					appendGroupCommentCpp[result,sys,defRule[[1]]]
				];
				
				AppendTo[result,
					{
						If[inHeader,
							{
								FMTCodeFormatter`noLineBreakT["extern"],
								FMTCodeFormatter`canTrimT[" "]
							},
							Nothing
						],
						sTypeName,
						FMTCodeFormatter`canTrimT[" "],
						FMTCodeFormatter`noLineBreakT[sObjName],
						FMTCodeFormatter`noLineBreakT[";"],
						FMTCodeFormatter`lineBreakT[]
					}
				]
			]
		],
		Normal[defs]
	];
	
	If[result!={},
		AppendTo[result,FMTCodeFormatter`verticalSpaceT[]]
	];
	
	result
]


(* Helper, generates operation's profile only *)
operationProfileCpp[sys_,inHeader_Symbol,pkgName_String,sOpName_String,profile_List,policies_Association]:=Module[
	{
		profileLen,
		count,
		result={}
	},
	
	(* Note: namespace qualification is added to operation's name at the point of definition,
		to ensure proper namespace association *)
	AppendTo[result,"void "<>If[inHeader,"",nameFromSymbolPolicyCpp[policies][pkgName,""]<>"::"]<>sOpName];
	
	profileLen=Length[profile];
	AppendTo[result,FMTCodeFormatter`noLineBreakT["("]];
	
	count=1;
	Scan[
		Function[paramDef, (* paramDef is mode[name\[Rule]type] *)
			With[
				{
					mode=Head[paramDef],
					sParamName=nameFromSymbolPolicyCpp[policies][paramDef[[1,1]],""],
					typeCat=FMTSystem`typeCategory[sys,paramDef[[1,2]]],
					sTypeName=nameFromSymbolPolicyCpp[policies][paramDef[[1,2]],pkgName]
				},
				
				AppendTo[result,
					{
						Switch[mode,
							FMTSymbols`in,
								If[typeCat===FMTSymbols`base||typeCat===FMTSymbols`enum||typeCat===FMTSymbols`array,
									(* simple types are passed by value,
										array types by degenerated pointer *)
									FMTCodeFormatter`noLineBreakT[sTypeName],
									
									(* structure types are passed by const reference *)
									{
										FMTCodeFormatter`noLineBreakT["const "<>sTypeName<>" &"],
										FMTCodeFormatter`canTrimT[" "],
										FMTCodeFormatter`noLineBreakT["/* in */"]
									}
								],
							
							FMTSymbols`inOut|FMTSymbols`out,
								If[typeCat===FMTSymbols`base||typeCat===FMTSymbols`enum||typeCat===FMTSymbols`structure,
									(* simple types and structures are passed by non-const reference *)
									{
										FMTCodeFormatter`noLineBreakT[sTypeName<>" &"],
										FMTCodeFormatter`canTrimT[" "],
										If[mode===FMTSymbols`inOut,
											FMTCodeFormatter`noLineBreakT["/* inOut */"],
											FMTCodeFormatter`noLineBreakT["/* out */"]
										]
									},
									
									(* array types are passed by degenerated pointer (no need for * here) *)
									{
										FMTCodeFormatter`noLineBreakT[sTypeName],
										FMTCodeFormatter`canTrimT[" "],
										If[mode===FMTSymbols`inOut,
											FMTCodeFormatter`noLineBreakT["/* inOut */"],
											FMTCodeFormatter`noLineBreakT["/* out */"]
										]
									}
								]
						],
						FMTCodeFormatter`canTrimT[" "],
						FMTCodeFormatter`noLineBreakT[sParamName],
						If[count<profileLen,
							{
								FMTCodeFormatter`noLineBreakT[","],
								FMTCodeFormatter`canTrimT[" "]
							},
							Nothing
						]
					}
				];
				++count
			]
		],
		profile
	];
	
	AppendTo[result,FMTCodeFormatter`noLineBreakT[")"]];
	
	result
]


operationSpecsCpp[sys_,pkgName_String,policies_Association]:=Module[
	{
		decls=FMTSystem`packageOperationDecls[sys,pkgName],
		result={}
	},
	
	If[Length[decls]!=0,
		result={
			"// Operation declarations.",
			FMTCodeFormatter`verticalSpaceT[]
		}
	];
	
	Scan[
		Function[declRule,
			With[
				{
					sOpName=nameFromSymbolPolicyCpp[policies][declRule[[1]],pkgName],
					profile=List@@declRule[[2]]
				},
				
				appendGroupCommentCpp[result,sys,declRule[[1]]];
				
				AppendTo[result,operationProfileCpp[sys,True,pkgName,sOpName,profile,policies]];
				
				AppendTo[result,
					{
						FMTCodeFormatter`noLineBreakT[";"],
						FMTCodeFormatter`lineBreakT[]
					}
				]
			]
		],
		Normal[decls]
	];
	
	AppendTo[result,FMTCodeFormatter`verticalSpaceT[]];
	
	result
]


(* Helper, generates code for a single object name
	or some part of it, if it is an array access or structure field access
*)
objectCodeCpp[obj_,pkgName_String,policies_Association]:=With[
	{
		wrapped=obj[[1]]
	},
	
	(* obj is objectT or objectLhsT or objectParentT,
		but it can wrap either a simple name (a symbol),
		or an array access or a structure field *)
	Switch[Head[wrapped],
		Symbol,
			FMTCodeFormatter`noLineBreakT[nameFromSymbolPolicyCpp[policies][wrapped,pkgName]],
		
		FMTSymbols`partT,
			partCodeCpp[wrapped,pkgName,policies],
		
		FMTSymbols`dotT,
			dotCodeCpp[wrapped,pkgName,policies]
	]
]

(* Helper, generates code for array access names *)
partCodeCpp[part_,pkgName_String,policies_Association]:=With[
	{
		parent=part[[1]],
		idx=part[[2]],
		indexOffset=policies[FMTOptions`arrayIndexOffset]
	},
	
	{
		objectCodeCpp[parent,pkgName,policies],
		FMTCodeFormatter`noLineBreakT["["],
		expressionCodeCpp[idx,Null,1,pkgName,policies],
		If[indexOffset!=0,
			{
				FMTCodeFormatter`canTrimT[" "],
				FMTCodeFormatter`noLineBreakT["-"],
				FMTCodeFormatter`canTrimT[" "],
				FMTCodeFormatter`noLineBreakT[ToString[indexOffset]]
			},
			
			Nothing
		],
		FMTCodeFormatter`noLineBreakT["]"]
	}
]

(* Helper, generates code for struct field names *)
dotCodeCpp[dot_,pkgName_String,policies_Association]:=With[
	{
		parent=dot[[1]],
		field=dot[[2]]
	},
	
	{
		objectCodeCpp[parent,pkgName,policies],
		FMTCodeFormatter`noLineBreakT["."],
		FMTCodeFormatter`noLineBreakT[nameFromSymbolPolicyCpp[policies][field,""]]
	}
]


(* Helper, generates code for a single literal *)
literalCodeCpp[lit_,outerExpr_Symbol,idx_,pkgName_String,policies_Association]:=Module[
	{
		value=lit[[1]],
		type=lit[[2]],
		result
	},
	
	Switch[type,
		FMTSymbols`integerT,
			If[value>=0,
				result=FMTCodeFormatter`noLineBreakT[ToString[value]],
				
				(* if value is negative and the containing expression is plusT,
					then instead of adding negative value, subtract the absolute value
					(plusCode already omitted the generation of plus symbol) *)
				Switch[outerExpr,
					FMTSymbols`plusT, (* for example x - 5, seen as plus[x,-5] *)
						If[idx==1,
							(* first value in addition is just a negative value, translate as is *)
							result=FMTCodeFormatter`noLineBreakT[ToString[value]],
							
							(* further values are subtractions instead of negations *)
							result={
								FMTCodeFormatter`noLineBreakT["-"],
								FMTCodeFormatter`canTrimT[" "],
								FMTCodeFormatter`noLineBreakT[ToString[-value]]
							}
						],
					
					FMTSymbols`timesT|FMTSymbols`divT, (* for example x * -5 or x / -5 *)
						If[idx==1,
							(* first value in multiplication or division is just a negative value,
								translate as is *)
							result=FMTCodeFormatter`noLineBreakT[ToString[value]],
							
							(* further values need parens *)
							result={
								FMTCodeFormatter`noLineBreakT["("],
								FMTCodeFormatter`noLineBreakT[ToString[value]],
								FMTCodeFormatter`noLineBreakT[")"]
							}
						],
					
					FMTSymbols`negT, (* for example x - -5 *)
						result={
							FMTCodeFormatter`noLineBreakT["("],
							FMTCodeFormatter`noLineBreakT[ToString[value]],
							FMTCodeFormatter`noLineBreakT[")"]
						},
					
					_,
						(* otherwise, translate the negative value as is *)
						result=FMTCodeFormatter`noLineBreakT[ToString[value]]
				]
			],
		
		FMTSymbols`booleanT,
			result=FMTCodeFormatter`noLineBreakT[If[value,"true","false"]],
		
		_,
			(* otherwise, it is an enum literal *)
			result=FMTCodeFormatter`noLineBreakT[nameFromSymbolPolicyCpp[policies][value,pkgName]]
	];
	
	result
]


(* Helper, generates code for a negated value (which is object or literal) *)
negCodeCpp[neg_,outerExpr_Symbol,idx_Integer,pkgName_String,policies_Association]:=Module[
	{
		value=neg[[1]],
		result
	},
	
	Switch[outerExpr,
		FMTSymbols`plusT, (* for example 5 - x, seen as plus[5,neg[x]] *)
			(* if the negated value is part of the plusT operation,
				it needs space between minus sign and the actual value,
				as the minus sign is generated as an operator (instead of plus),
				not a negator
				but only for second and later elements of the plus operation
			*)
			result={
				FMTCodeFormatter`noLineBreakT["-"],
				If[idx>1,
					FMTCodeFormatter`canTrimT[" "],
					Nothing
				],
				expressionCodeCpp[value,FMTSymbols`negT,1,pkgName,policies]
			},
		
		FMTSymbols`negT, (* multiple negation is allowed, but needs parens *)
			result={
				FMTCodeFormatter`noLineBreakT["(-"],
				expressionCodeCpp[value,FMTSymbols`negT,1,pkgName,policies],
				FMTCodeFormatter`noLineBreakT[")"]
			},
		
		_, (* regular negation *)
			result={
				FMTCodeFormatter`noLineBreakT["-"],
				expressionCodeCpp[value,FMTSymbols`negT,1,pkgName,policies]
			}
	];
	
	result
]


(* Helper, generates code for a plus operator *)
plusCodeCpp[plus_,pkgName_String,policies_Association]:=Module[
	{
		len=Length[plus],
		count,
		result={}
	},
	
	count=1;
	Scan[
		Function[part, (* part is a nested expression *)
			(* recursively translate all involved parts *)
			AppendTo[result,expressionCodeCpp[part,FMTSymbols`plusT,count,pkgName,policies]];
			
			If[count<len,
				If[Head[plus[[count+1]]]===FMTSymbols`negT||
					(Head[plus[[count+1]]]===FMTSymbols`literalT&&
						plus[[count+1,2]]===FMTSymbols`integerT&&
						plus[[count+1,1]]<0),
					
					(* the next element is either negated or it is a negative integer literal
						\[Rule] do not generate any operator sign, negCode and literalCode will
						generate the minus sign instead *)
					AppendTo[result,FMTCodeFormatter`canTrimT[" "]],
					
					(* use plus sign for regular additions *)
					AppendTo[result,
						{
							FMTCodeFormatter`canTrimT[" "],
							FMTCodeFormatter`noLineBreakT["+"],
							FMTCodeFormatter`canTrimT[" "]
						}
					]
				]
			];
			
			++count
		],
		plus
	];
	
	result
]


(* Helper, generates code for a times operator *)
timesCodeCpp[times_,pkgName_String,policies_Association]:=Module[
	{
		len=Length[times],
		count,
		result={}
	},
	
	count=1;
	Scan[
		Function[part, (* part is a nested expression *)
			(* recursively translate all involved parts *)
			AppendTo[result,expressionCodeCpp[part,FMTSymbols`timesT,count,pkgName,policies]];
			
			If[count<len,
				AppendTo[result,
					{
						FMTCodeFormatter`canTrimT[" "],
						FMTCodeFormatter`noLineBreakT["*"],
						FMTCodeFormatter`canTrimT[" "]
					}
				]
			];
			
			++count
		],
		times
	];
	
	result
]


(* Helper, generates code for a div operator *)
divCodeCpp[div_,pkgName_String,policies_Association]:=Module[
	{
		len=Length[div],
		count,
		result={}
	},
	
	count=1;
	Scan[
		Function[part, (* part is a nested expression *)
			(* recursively translate all involved parts *)
			AppendTo[result,expressionCodeCpp[part,FMTSymbols`divT,count,pkgName,policies]];
			
			If[count<len,
				AppendTo[result,
					{
						FMTCodeFormatter`canTrimT[" "],
						FMTCodeFormatter`noLineBreakT["/"],
						FMTCodeFormatter`canTrimT[" "]
					}
				]
			];
			
			++count
		],
		div
	];
	
	result
]


(* Helper, generates code for a generic operator: and, or, \[Equal], \[NotEqual], <, \[LessEqual], >, \[GreaterEqual] *)
generalOperatorCodeCpp[op_,pkgName_String,policies_Association]:=Module[
	{
		len=Length[op],
		count,
		result={}
	},
	
	count=1;
	Scan[
		Function[part, (* part is a nested expression *)
			(* recursively translate all involved parts *)
			AppendTo[result,expressionCodeCpp[part,Head[op],count,pkgName,policies]];
			
			If[count<len,
				AppendTo[result,
					{
						FMTCodeFormatter`canTrimT[" "],
						Switch[Head[op],
							FMTSymbols`andT,
								FMTCodeFormatter`noLineBreakT["&&"],
							
							FMTSymbols`orT,
								FMTCodeFormatter`noLineBreakT["||"],
							
							FMTSymbols`equalT,
								FMTCodeFormatter`noLineBreakT["=="],
							
							FMTSymbols`unequalT,
								FMTCodeFormatter`noLineBreakT["!="],
							
							FMTSymbols`lessT,
								FMTCodeFormatter`noLineBreakT["<"],
							
							FMTSymbols`lessEqualT,
								FMTCodeFormatter`noLineBreakT["<="],
							
							FMTSymbols`greaterT,
								FMTCodeFormatter`noLineBreakT[">"],
							
							FMTSymbols`greaterEqualT,
								FMTCodeFormatter`noLineBreakT[">="]
						],
						FMTCodeFormatter`canTrimT[" "]
					}
				]
			];
			
			++count
		],
		op
	];
	
	result
]


(* Helper, generates code for a not operator *)
notCodeCpp[not_,pkgName_String,policies_Association]:={
	FMTCodeFormatter`noLineBreakT["!"], (* no line breaks here *)
	expressionCodeCpp[not[[1]],FMTSymbols`notT,1,pkgName,policies]
}


(* Helper, generates code for the whole expression,
	outerExpr is a Head of the outer (containing) expression (or Null),
	idx is an index within the outer expression,
	both are used to decide whether negation sign needs space or whether parens are needed *)
expressionCodeCpp[expr_,outerExpr_Symbol,idx_Integer,pkgName_String,policies_Association]:=Module[
	{
		result
	},
	
	Switch[Head[expr],
		FMTSymbols`objectT,
			result=objectCodeCpp[expr,pkgName,policies],
		
		FMTSymbols`literalT,
			result=literalCodeCpp[expr,outerExpr,idx,pkgName,policies],
		
		FMTSymbols`negT,
			result=negCodeCpp[expr,outerExpr,idx,pkgName,policies],
		
		FMTSymbols`plusT,
			If[MemberQ[{FMTSymbols`timesT,FMTSymbols`divT,FMTSymbols`negT},outerExpr],
				(* parens are needed for the whole plus operation *)
				result={
					FMTCodeFormatter`noLineBreakT["("],
					plusCodeCpp[expr,pkgName,policies],
					FMTCodeFormatter`noLineBreakT[")"]
				},
				
				(* no parens needed *)
				result=plusCodeCpp[expr,pkgName,policies]
			],
		
		FMTSymbols`timesT,
			If[outerExpr===FMTSymbols`negT,
				(* parens are needed for the whole times operation *)
				result={
					FMTCodeFormatter`noLineBreakT["("],
					timesCodeCpp[expr,pkgName,policies],
					FMTCodeFormatter`noLineBreakT[")"]
				},
				
				(* no parens needed *)
				result=timesCodeCpp[expr,pkgName,policies]
			],
		
		FMTSymbols`divT,
			If[outerExpr===FMTSymbols`negT,
				(* parens are needed for the whole div operation *)
				result={
					FMTCodeFormatter`noLineBreakT["("],
					divCodeCpp[expr,pkgName,policies],
					FMTCodeFormatter`noLineBreakT[")"]
				},
				
				(* no parens needed *)
				result=divCodeCpp[expr,pkgName,policies]
			],
		
		FMTSymbols`andT|FMTSymbols`orT|
		FMTSymbols`equalT|FMTSymbols`unequalT|
		FMTSymbols`lessT|FMTSymbols`lessEqualT|
		FMTSymbols`greaterT|FMTSymbols`greaterEqualT,
			If[outerExpr=!=Null,
				(* just in case, parens are needed for the whole operation *)
				result={
					FMTCodeFormatter`noLineBreakT["("],
					generalOperatorCodeCpp[expr,pkgName,policies],
					FMTCodeFormatter`noLineBreakT[")"]
				},
				
				(* no parens needed *)
				result=generalOperatorCodeCpp[expr,pkgName,policies]
			],
		
		FMTSymbols`notT,
			result=notCodeCpp[expr,pkgName,policies]
	];
	
	result
]


(* Helper statement-related code generators, separate functions for each statement type *)

userCommentCpp[comment_FMTSymbols`commentT]:=With[
	{
		lines=List@@comment
	},
	
	Map[
		Function[line,
			{
				"// "<>line,
				FMTCodeFormatter`lineBreakT[]
			}
		],
		lines
	]
]

appendGroupCommentCpp[result_,sys_,entityName_]:=With[
	{
		comment=FMTSystem`fmtGetComment[sys,entityName]
	},
	
	If[Head[comment]===FMTSymbols`commentT,
		AppendTo[result,FMTCodeFormatter`verticalSpaceT[]];
		AppendTo[result,userCommentCpp[comment]],
		
		If[comment===Null,
			(* this means that there is no comment,
				but the entity is the first in its group *)
			AppendTo[result,FMTCodeFormatter`verticalSpaceT[]]
		]
	]
]

SetAttributes[appendGroupCommentCpp,HoldFirst]

emptyStatementCodeCpp[]:={}

compoundStatementCodeCpp[comp_FMTSymbols`compoundT,pkgName_String,policies_Association]:=Module[
	{
		len=Length[comp],
		count,
		addVerticalSpaceBefore,
		addVerticalSpaceAfter,
		result={}
	},
	
	count=1;
	Scan[
		Function[st, (* st is a single statement within a compound *)
			(* add vertical space around control structures and procedure calls *)
			If[MemberQ[{FMTSymbols`ifT,FMTSymbols`whileT,FMTSymbols`switchT,FMTSymbols`callT},
					Head[st]],
					
				addVerticalSpaceBefore=1<count;
				addVerticalSpaceAfter=count<len
			];
			++count;
			
			If[Head[st]=!=FMTSymbols`statementT, (* avoid empty statements in compounds *)
				If[addVerticalSpaceBefore,
					AppendTo[result,FMTCodeFormatter`verticalSpaceT[]]
				];
				
				AppendTo[result,bodyCodeFromStatementCpp[st,pkgName,policies]];
				
				If[addVerticalSpaceAfter,
					AppendTo[result,FMTCodeFormatter`verticalSpaceT[]]
				]
			]
		],
		comp
	];
	
	result
]

assignmentCodeCpp[st_FMTSymbols`assignmentT,pkgName_String,policies_Association]:=With[
	{
		lhsObject=st[[1]],
		rhsExpression=st[[2]]
	},
	
	{
		objectCodeCpp[lhsObject,pkgName,policies],
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT["="],
		FMTCodeFormatter`canTrimT[" "],
		expressionCodeCpp[rhsExpression,Null,1,pkgName,policies],
		FMTCodeFormatter`noLineBreakT[";"],
		FMTCodeFormatter`lineBreakT[]
	}
]

(* this is used to implement copy of whole-array objects *)
arrayCopyCpp[st_arrayCopyCppT,pkgName_String,policies_Association]:=With[
	{
		qArrayType=st[[1]],
		qElementType=st[[2]],
		arraySize=st[[3]],
		lhsObject=st[[4]],
		rhsExpression=st[[5]],
		memCpy=memCpyPolicyCpp[policies]
	},
	
	(* array copy operation is generated by a separate function,
		which might be overriden by means of policies or options *)
	memCpy[
		objectCodeCpp[lhsObject,pkgName,policies],
		expressionCodeCpp[rhsExpression,Null,1,pkgName,policies],
		nameFromSymbolPolicyCpp[policies][qArrayType,pkgName],
		nameFromSymbolPolicyCpp[policies][qElementType,pkgName],
		arraySize
	]
]

ifCodeCpp[st_FMTSymbols`ifT,pkgName_String,policies_Association]:=With[
	{
		cond=st[[1]],
		thenBranch=st[[2]],
		elseBranch=If[Length[st]>2,st[[3]],Null]
	},
	
	{
		FMTCodeFormatter`noLineBreakT["if"],
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT["("],
		expressionCodeCpp[cond,Null,1,pkgName,policies],
		FMTCodeFormatter`noLineBreakT[")"],
		FMTCodeFormatter`lineBreakT[],
		FMTCodeFormatter`noLineBreakT["{"],
		FMTCodeFormatter`lineBreakT[],
		FMTCodeFormatter`moreIndentT[],
		bodyCodeFromStatementCpp[thenBranch,pkgName,policies],
		FMTCodeFormatter`lessIndentT[],
		If[elseBranch=!=Null,
			{
				FMTCodeFormatter`noLineBreakT["}"],
				FMTCodeFormatter`lineBreakT[],
				FMTCodeFormatter`noLineBreakT["else"],
				FMTCodeFormatter`lineBreakT[],
				FMTCodeFormatter`noLineBreakT["{"],
				FMTCodeFormatter`lineBreakT[],
				FMTCodeFormatter`moreIndentT[],
				bodyCodeFromStatementCpp[elseBranch,pkgName,policies],
				FMTCodeFormatter`lessIndentT[]
			},
			Nothing
		],
		FMTCodeFormatter`noLineBreakT["}"],
		FMTCodeFormatter`lineBreakT[]
	}
]

whileCodeCpp[while_FMTSymbols`whileT,pkgName_String,policies_Association]:=With[
	{
		cond=while[[1]],
		loopBranch=while[[2]]
	},
	
	{
		FMTCodeFormatter`noLineBreakT["while"],
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT["("],
		expressionCodeCpp[cond,Null,1,pkgName,policies],
		FMTCodeFormatter`noLineBreakT[")"],
		FMTCodeFormatter`lineBreakT[],
		FMTCodeFormatter`noLineBreakT["{"],
		FMTCodeFormatter`lineBreakT[],
		FMTCodeFormatter`moreIndentT[],
		bodyCodeFromStatementCpp[loopBranch,pkgName,policies],
		FMTCodeFormatter`lessIndentT[],
		FMTCodeFormatter`noLineBreakT["}"],
		FMTCodeFormatter`lineBreakT[]
	}
]

caseValuesCpp[values_,pkgName_String,policies_Association]:=
	If[values===FMTSymbols`defaultT[],
		{
			FMTCodeFormatter`noLineBreakT["default:"],
			FMTCodeFormatter`lineBreakT[]
		},
		
		If[Head[values]===FMTSymbols`literalT,
			(* it is a single value *)
			{
				FMTCodeFormatter`noLineBreakT["case "<>nameFromSymbolPolicyCpp[policies][values[[1]],pkgName]<>":"],
				FMTCodeFormatter`lineBreakT[]
			},
			
			(* it is a list of alternatives, each of them is translated into a separate "case value:" label *)
			Map[
				Function[v, (* v is a single literalT[enumName,enumType] from the list of alternatives *)
					{
						FMTCodeFormatter`noLineBreakT["case "<>nameFromSymbolPolicyCpp[policies][v[[1]],pkgName]<>":"],
						FMTCodeFormatter`lineBreakT[]
					}
				],
				List@@values
			]
		]
	]

switchCodeCpp[switch_FMTSymbols`switchT,pkgName_String,policies_Association]:=Module[
	{
		expr=switch[[1]],
		result
	},
	
	result={
		FMTCodeFormatter`noLineBreakT["switch"],
		FMTCodeFormatter`canTrimT[" "],
		FMTCodeFormatter`noLineBreakT["("],
		expressionCodeCpp[expr,Null,1,pkgName,policies],
		FMTCodeFormatter`noLineBreakT[")"],
		FMTCodeFormatter`lineBreakT[],
		FMTCodeFormatter`noLineBreakT["{"],
		FMTCodeFormatter`lineBreakT[]
	};
	
	AppendTo[result,
		Riffle[
			Map[
				Function[vb, (* vb is {values,branch} *)
					With[
						{
							values=vb[[1]],
							branch=vb[[2]]
						},
						
						{
							caseValuesCpp[values,pkgName,policies],
							FMTCodeFormatter`lineBreakT[],
							FMTCodeFormatter`moreIndentT[],
							bodyCodeFromStatementCpp[branch,pkgName,policies],
							FMTCodeFormatter`noLineBreakT["break;"],
							FMTCodeFormatter`lineBreakT[],
							FMTCodeFormatter`lessIndentT[]
						}
					]
				],
				Partition[List@@switch[[2;;]],2]
			],
			FMTCodeFormatter`verticalSpaceT[]
		]
	];
	
	AppendTo[result,
		{
			FMTCodeFormatter`noLineBreakT["}"],
			FMTCodeFormatter`lineBreakT[]
		}
	];
	
	result
]

callCodeCpp[call_FMTSymbols`callT,pkgName_String,policies_Association]:=Module[
	{
		qProcName=call[[1]],
		formals=call[[2]],
		actuals=call[[3]],
		len,
		count,
		result
	},
	
	len=Length[actuals];
	
	result={FMTCodeFormatter`noLineBreakT[nameFromSymbolPolicyCpp[policies][qProcName,pkgName]]};
	
	AppendTo[result,FMTCodeFormatter`noLineBreakT["("]];
	
	count=1;
	Scan[
		Function[actual,
			AppendTo[result,expressionCodeCpp[actual,Null,1,pkgName,policies]];
			
			If[count<len,
				AppendTo[result,
					{
						FMTCodeFormatter`noLineBreakT[","],
						FMTCodeFormatter`canTrimT[" "]
					}
				]
			];
			
			++count
		],
		actuals
	];
	
	AppendTo[result,
		{
			FMTCodeFormatter`noLineBreakT[")"],
			FMTCodeFormatter`noLineBreakT[";"],
			FMTCodeFormatter`lineBreakT[]
		}
	];
	
	result
]

bodyCodeFromStatementCpp[st_,pkgName_String,policies_Association]:=Module[
	{
		result
	},
	
	Switch[Head[st],
		FMTSymbols`commentT,
			(* this is a user-injected comment *)
			result=userCommentCpp[st],
		
		FMTSymbols`statementT,
			(* this is a Null statement (perhaps a null procedure) *)
			result=emptyStatementCodeCpp[],
		
		FMTSymbols`compoundT,
			(* this is a compound statement (a sequence of statements) *)
			result=compoundStatementCodeCpp[st,pkgName,policies],
		
		FMTSymbols`assignmentT,
			result=assignmentCodeCpp[st,pkgName,policies],
		
		arrayCopyCppT,
			result=arrayCopyCpp[st,pkgName,policies],
		
		FMTSymbols`ifT,
			result=ifCodeCpp[st,pkgName,policies],
		
		FMTSymbols`whileT,
			result=whileCodeCpp[st,pkgName,policies],
		
		FMTSymbols`switchT,
			result=switchCodeCpp[st,pkgName,policies],
		
		FMTSymbols`callT,
			result=callCodeCpp[st,pkgName,policies]
	];
	
	result
]


(* Helper, prepares the operation body by:
	- wrapping assignments of array objects in arrayCopyCppT[],
		so that they are recognized and implemented in terms of memcpy
*)
preprocessBodyCpp[sys_,profile_,body_]:=Module[
	{
		arrayTypes=FMTSystem`typesForCategory[sys,FMTSymbols`array],
		replacements={}
	},
	
	(* gather the list of array types that are used in assignment operations *)
	Scan[
		Function[at, (* array type, typeName\[Rule]array[elemType,size] *)
			With[
				{
					qArrayType=at[[1]],
					qElementType=at[[2,1]],
					arraySize=at[[2,2]]
				},
				
				AppendTo[replacements,
					FMTSymbols`assignmentT[FMTSymbols`objectLhsT[n_,qArrayType],rhs_]->
						arrayCopyCppT[
							qArrayType,
							qElementType,
							arraySize,
							FMTSymbols`objectLhsT[n,qArrayType],
							rhs
						]
				]
			]
		],
		arrayTypes
	];
	
	body/.replacements
]


operationDefsCpp[sys_,pkgName_String,policies_Association]:=Module[
	{
		decls=FMTSystem`packageOperationDecls[sys,pkgName],
		result={}
	},
	
	Scan[
		Function[declRule,
			With[
				{
					qOpName=declRule[[1]],
					sOpName=nameFromSymbolPolicyCpp[policies][declRule[[1]],pkgName],
					profile=List@@declRule[[2]],
					localVars=FMTSystem`operationLocalVars[sys,declRule[[1]]],
					opBody=preprocessBodyCpp[
						sys,
						declRule[[2]],
						FMTSystem`resolvedOperationBody[sys,pkgName,declRule[[1]],False,{}]
					]
				},
				
				If[opBody=!=Null,
					If[result!={},
						AppendTo[result,FMTCodeFormatter`verticalSpaceT[]]
					];
					
					AppendTo[result,
						{
							operationProfileCpp[sys,False,pkgName,sOpName,profile,policies],
							FMTCodeFormatter`lineBreakT[],
							FMTCodeFormatter`noLineBreakT["{"],
							FMTCodeFormatter`lineBreakT[],
							FMTCodeFormatter`moreIndentT[]
						}
					];
					
					(* define local variables *)
					If[localVars!={},
						Scan[
							Function[varDef, (* varDef is name\[Rule]type *)
								With[
									{
										name=nameFromSymbolPolicyCpp[policies][varDef[[1]],pkgName],
										type=nameFromSymbolPolicyCpp[policies][varDef[[2]],pkgName]
									},
									
									AppendTo[result,
										{
											type,
											FMTCodeFormatter`canTrimT[" "],
											FMTCodeFormatter`noLineBreakT[name],
											FMTCodeFormatter`noLineBreakT[";"],
											FMTCodeFormatter`lineBreakT[]
										}
									]
								]
							],
							localVars
						];
						AppendTo[result,FMTCodeFormatter`verticalSpaceT[]];
					];
					
					(* generate procedure's body *)
					
					AppendTo[result,bodyCodeFromStatementCpp[opBody,pkgName,policies]];
					
					AppendTo[result,
						{
							FMTCodeFormatter`lessIndentT[],
							"}"
						}
					]
				]
			]
		],
		Normal[decls]
	];
	
	If[result!={},
		{
			"// Operation definitions.",
			FMTCodeFormatter`verticalSpaceT[],
			
			result,
			FMTCodeFormatter`lineBreakT[]
		},
		
		result
	]
]


(* Helper, generates #includes for needed standard headers *)
standardHeadersCpp[tokens_,policies_Association]:=Module[
	{
		allStringTokens=Cases[tokens,_String,All],
		needStdInt,
		needString
	},
	
	needStdInt=MemberQ[allStringTokens,integerPolicyCpp[policies]];
	needString=MemberQ[allStringTokens,"std::memmove"];
	If[needStdInt||needString,
		{
			If[needStdInt,
				{
					"#include <cstdint>",
					FMTCodeFormatter`lineBreakT[]
				},
				Nothing
			],
			If[needString,
				{
					"#include <cstring>",
					FMTCodeFormatter`lineBreakT[]
				},
				Nothing
			],
			FMTCodeFormatter`verticalSpaceT[]
		},
		Nothing
	]
]


generateSpecCpp[sys_,pkgName_String,fileName_String,specDeps_List,hasBody_,policies_Association]:=Module[
	{
		sPkgName=nameFromSymbolPolicyCpp[policies][pkgName,""],
		indentSize=indentSizePolicyCpp[policies],
		maxLineLength=maxLineLengthPolicyCpp[policies],
		preamble=preamblePolicyCpp[policies],
		pkgComment,
		headerGuard,
		types,
		objects,
		operations
	},
	
	headerGuard="FILE_"<>ToUpperCase[sPkgName]<>"_H_INCLUDED";
	
	pkgComment=FMTSystem`fmtGetComment[sys,pkgName];
	
	(* generate type definitions *)
	types=typeDefinitionsCpp[sys,pkgName,policies];
	
	(* generate object declarations *)
	(* Note: object definitions are generated before operations,
		because operation profiles can refer to objects in their pre- and post-conditions.
		This means that objects (at least those that are referred) need to be defined
		first in the model and this convention is preserved in the generated code.
	*)
	objects=objectSpecsCpp[sys,True,pkgName,policies]; (* True - header declarations, without definitions *)
	
	(* generate operation prototypes *)
	operations=operationSpecsCpp[sys,pkgName,policies];
	
	FMTCodeFormatter`format[Flatten@{
		(* generate file preamble, if requested by policies (by default empty) *)
		preamble[pkgName,sPkgName,fileName],
		
		(* generate header guards *)
		"#ifndef "<>headerGuard,
		FMTCodeFormatter`lineBreakT[],
		"#define "<>headerGuard,
		FMTCodeFormatter`lineBreakT[],
		FMTCodeFormatter`verticalSpaceT[],
		
		(* generate includes for dependent packages *)
		includesCpp[specDeps,policies],
		
		FMTCodeFormatter`verticalSpaceT[],
		
		standardHeadersCpp[{types,objects,operations},policies],
		
		If[Head[pkgComment]===FMTSymbols`commentT,
			{
				userCommentCpp[pkgComment]
			},
			Nothing
		],
		
		"namespace "<>sPkgName,
		FMTCodeFormatter`lineBreakT[],
		FMTCodeFormatter`noLineBreakT["{"],
		FMTCodeFormatter`lineBreakT[],
		(*FMTCodeFormatter`moreIndentT[],*)
		FMTCodeFormatter`verticalSpaceT[],
		
		If[Not[hasBody],
			{
				"// This package does not need body.",
				FMTCodeFormatter`lineBreakT[],
				FMTCodeFormatter`verticalSpaceT[]
			},
			
			Nothing
		],
		
		types,
		objects,
		operations,
		
		FMTCodeFormatter`verticalSpaceT[],
		(*FMTCodeFormatter`lessIndentT[],*)
		"} // namespace "<>sPkgName,
		FMTCodeFormatter`lineBreakT[],
		FMTCodeFormatter`verticalSpaceT[],
		"#endif // "<>headerGuard,
		FMTCodeFormatter`lineBreakT[]
	},
	<|
		FMTCodeFormatter`indentSize->indentSize,
		FMTCodeFormatter`maxLineLength->maxLineLength
	|>]
]


generateBodyCpp[sys_,pkgName_String,fileName_String,bodyDeps_List,policies_Association]:=Module[
	{
		sPkgName=nameFromSymbolPolicyCpp[policies][pkgName,""],
		indentSize=indentSizePolicyCpp[policies],
		maxLineLength=maxLineLengthPolicyCpp[policies],
		preamble=preamblePolicyCpp[policies],
		variables,
		operations,
		result
	},
	
	(* generate variable definitions *)
	variables=objectSpecsCpp[sys,False,pkgName,policies]; (* False - body definitions *)
	
	(* generate operation definitions *)
	operations=operationDefsCpp[sys,pkgName,policies];
	
	If[variables!={}||operations!={},
		result=FMTCodeFormatter`format[Flatten@{
			(* generate file preamble, if requested by policies (by default empty) *)
			preamble[pkgName,sPkgName,fileName],
		
			(* generate includes for dependent packages,
				the current package is taken care of as well *)
			includesCpp[DeleteDuplicates[Join[{pkgName},bodyDeps]],policies],
			
			FMTCodeFormatter`verticalSpaceT[],
			
			standardHeadersCpp[{variables,operations},policies],
			
			"using namespace "<>sPkgName<>";",
			FMTCodeFormatter`lineBreakT[],
			FMTCodeFormatter`verticalSpaceT[],
			
			variables,
			operations
		},
		<|
			FMTCodeFormatter`indentSize->indentSize,
			FMTCodeFormatter`maxLineLength->maxLineLength
		|>],
		
		result=Null
	];
	
	result
]


fmtGenerateCpp[sys_,pkgName_String,policies_Association]:=Module[
	{
		fNames,
		dependencies=FMTSystem`dependentPackages[sys,pkgName],
		spec,
		body,
		result=<||>
	},
	
	fNames=fileNamesCpp[pkgName,policies];
	
	body=generateBodyCpp[sys,pkgName,fNames[[2]],dependencies[[2]],policies];
	spec=generateSpecCpp[sys,pkgName,fNames[[1]],dependencies[[1]],body=!=Null,policies];
	
	result[fNames[[1]]]=spec;
	result[fNames[[2]]]=body;
	
	result
]


End[];


EndPackage[];
