(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


Needs["FMTSymbols`"];


BeginPackage["FMTOptions`"];


(* This association contains default options that influence the behaviour in various aspects of FMT.
	The options can be read and overwritten by the user and passed to the system constructor function (createSoftwareSystem) *)
defaultOptions=<|
	(* if True, interactive buttons with further actions are shown
		after introducing new or replacing an old operation definition;
		set this to False for scripted sessions or bulk processing *)
	showInteractiveButtons->True,
	
	(* range of standard integer type, inclusive, as used in proofs *)
	integerBounds->{-2^31,2^31-1},
	
	(* function used for rounding results of integer division, as used in proofs *)
	integerRoundingFunction->IntegerPart,
	
	(* index of the first array element, used in proofs and in generated code;
		0 is a conventional value for C and C++ generated code,
		1 might be a popular choice for Ada *)
	arrayIndexOffset->0,
	
	(* preprocessor function that is run on the operation's body
		at the beginning of the syntax resolution process;
		the default is implemented below and is an example of For\[Rule]While translation *)
	grammarPreprocessor->defaultPreprocessGrammar,
	
	(* formatter for printing expressions
		- used for knowns and verification conditions in proof statements *)
	formulaFormatter->StandardForm
|>;


(* Example grammar preprocessor that implements For loop in terms of While construct *)
defaultPreprocessGrammar[expression_]:=Module[
	{
		processed
	},
	
	(* implement simple translation For \[Rule] While *)
	
	(* Note: intermediate translation via forT is needed,
		because For is a built-in symbol and using it directly in the replacement rule
		causes troubles, as the language tries to process it according to its
		built-in meaning, other symbols do not require such treatment.
	*)
	processed=expression/.For->FMTSymbols`forT;
	
	(* FixedPoint allows to handle the case of nested For loops,
		where a single replacement would cover only the external loop
	*)
	processed=FixedPoint[
		Function[ex,
			ex/.FMTSymbols`forT[init_,cond_,step_,body_]->
			FMTSymbols`compoundT[
				init,
				FMTSymbols`whileT[cond,
					FMTSymbols`compoundT[body,step]
				]
			]
		],
		processed
	];
	
	processed
]


EndPackage[];
