(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


Needs["FMTProofsAggregates`"];
Needs["FMTProofsBranches`"];
Needs["FMTProofsFreezings`"];
Needs["FMTSymbols`"];
Needs["FMTSystem`"];
Needs["FMTVerification`"];


BeginPackage["FMTProofs`"];


proofStatements::usage="proofStatements[sys,packageName,qOperationName,variableDomains] Returns the list of proof statements (not yet verified) for the given operation\n"<>
"and fills the domain information for all used variables.\n"<>
"Exposed in the public interface as an inspection aid.";


Begin["Private`"];


(* Helper, iterates over all numberedT elements (at the top-level) of the given subexpression
	and accumulates path statements for them *)
iterateNumbered[
	sys_,
	packageName_,
	qOperationName_,
	opBody_,
	allPathsStatements_,
	lastCumulatedStatements_,
	branchContext_,
	cumulationWrapperFunction_,
	freezingRenames_,
	lastFreezingIds_,
	variableDomains_,
	cumulatedAggregateAccesses_,
	expr_
]:=
	Cases[expr,
		FMTSymbols`numberedT[numIndex_,numExpr_]
		:>
		processSingleNumbered[
			sys,
			packageName,
			qOperationName,
			opBody,
			allPathsStatements,
			lastCumulatedStatements,
			branchContext,
			cumulationWrapperFunction,
			freezingRenames,
			lastFreezingIds,
			variableDomains,
			cumulatedAggregateAccesses,
			numIndex,
			numExpr
		],
		
		If[Head[expr]===FMTSymbols`numberedT,
			{0},
			{1}
		]
	]

SetAttributes[iterateNumbered,HoldAll]


(* Helper, reverts symbol replacements within oldT wrappers,
	freezings is a list of freezing to apply to oldT elements,
	usually empty except for operation calls, where it is equal
	to the list of freezings from the point right before the call,
	so that oldT elements (in operation post-conditions)
	can properly refer to their respective initial values *)
restoreOldValues[expression_,freezings_]:=Module[
	{
		result,
		oldTPositions=Position[expression,_FMTSymbols`oldT]
	},
	
	(* replace oldT[newSymbol] with oldT[oldSymbol],
		taking the given list of freezings into account: *)
	result=MapAt[
		Function[part,
			FMTProofsFreezings`unfreeze[part]/.freezings
		],
		expression,
		oldTPositions
	];
	
	(* finally, remove the wrappers themselves: *)
	result=result/.FMTSymbols`oldT[subExpr_]->subExpr;
	
	result
]


(* Helper, appends wrapped knowns to the list of cumulated knowns *)
appendWrappedKnowns[knowns_,lastCumulatedStatements_,cumulationWrapperFunction_]:=
	If[knowns!={},
		Scan[
			Function[known,
				AppendTo[lastCumulatedStatements,
					cumulationWrapperFunction[known]
				]
			],
			knowns
		]
	]

SetAttributes[appendWrappedKnowns,HoldRest]

(* Helper, performs recursive calls for If/While/Switch branches and generates alternative (conditional) assignments
	for those variables that have changed within those branches *)
processIfWhileSwitch[
	sys_,
	packageName_,
	qOperationName_,
	opBody_,
	allPathsStatements_,
	lastCumulatedStatements_,
	branchContext_,
	cumulationWrapperFunction_,
	freezingRenames_,
	lastFreezingIds_,
	variableDomains_,
	cumulatedAggregateAccesses_,
	expr_
]:=Module[
	{
		branchesToProcess,
		conditionExpression,
		switchLabelsUsed,
		branchConditions,
		collectStrongRenamings,
		collectWeakRenamings,
		frozenKnowns,
		aggregateKnowns,
		aggregateAccessPreKnownPropagations, (* separate knowns related to array indexing and struct field accesses *)
		aggregateAccessPostKnownAssignments, (* note: there are no aggregateAccessPreKnownAssignments here, *)
		aggregateAccessPostKnownPropagations, (* because PREconditions do not generate assignments,
												and POST values include also knowns from the operation itself *)
		rememberedLastCumulatedStatements,
		branchCumulatedStatements, (* increments of lastCumulatedStatements in each branch *)
		rememberedCumulatedAggregateAccesses,
		branchCumulatedAggregateAccesses, (* increments of cumulatedAggregateAccesses in each branch *)
		rememberedFreezingRenames,
		conditionalFreezingUpdates, (* <| condition\[Rule]<| name\[Rule]freezeRenaming, ... |>, ... |> *)
		conditionalRenamingAssignments,
		condAssignments,
		localVariablesThatNeedInitialization,
		sourceObject,
		objectToRename,
		objectsBranchRenamed={}, (* set of objects that got new renamings after branch closing *)
		options=FMTSystem`fmtGetCurrentOptions[sys]
	},
	
	(* recurse on branches: *)
	
	(* Note: collect strong renamings only for If and Switch statements,
		While statements are considered to be black-boxes
		and only loop invariants can be used to export
		some knowledge from inside to outside of the loops
	*)
	Switch[Head[expr],
		FMTSymbols`ifT,
			branchesToProcess=List@@expr[[2;;]]; (* "then" branch with "else", if present *)
			conditionExpression=FMTProofsFreezings`unwrapFrozenAndGrammar[
				FMTProofsFreezings`applyCurrentFreezings[expr[[1]],freezingRenames],
				options
			];
			
			(* consider both "then" and "else" conditions, even if the "else"
				branch does not explicitly exist - this is to enable the creation
				of "weak" variable assignments, which in turn allow values to be traced
				also outside of branches, when they are modified
			*)
			branchConditions={conditionExpression,Not[conditionExpression]};
			collectStrongRenamings=True;
			collectWeakRenamings=True,
		
		FMTSymbols`switchT,
			branchesToProcess=List@@expr[[3;; ;;2]]; (* 1 is enum object,
													2,4,6,... are switch labels,
													3,5,7,... are actual branches *)
			switchLabelsUsed={};
			branchConditions=Map[
				Function[caseLabel, (* caseLabel is a single enum value or a list of alternatives,
										or a blank
										in either case it is an expression that involves literalT,
										and if no literalT is present, then it was a blank *)
					With[
						{
							labels=Cases[caseLabel,_FMTSymbols`literalT,All]
						},
						
						If[labels!={},
							AppendTo[switchLabelsUsed,labels];
							Or@@Map[
								Function[singleLabel,
									FMTProofsFreezings`unwrapFrozenAndGrammar[
										FMTProofsFreezings`applyCurrentFreezings[
											FMTSymbols`equalT[expr[[1]],singleLabel],
											freezingRenames
										],
										options
									]
								],
								labels
							],
							
							(* there were no literals found in this case label,
								this must be a blank *)
							And@@Map[
								Function[singleLabel,
									FMTProofsFreezings`unwrapFrozenAndGrammar[
										FMTProofsFreezings`applyCurrentFreezings[
											FMTSymbols`unequalT[expr[[1]],singleLabel],
											freezingRenames
										],
										options
									]
								],
								Flatten[switchLabelsUsed]
							]
						]
					]
				],
				List@@expr[[2;; ;;2]]
			];
			collectStrongRenamings=True;
			collectWeakRenamings=True,
		
		FMTSymbols`whileT,
			branchesToProcess={expr[[2]]}; (* loop body *)
			conditionExpression=FMTProofsFreezings`unwrapFrozenAndGrammar[
				FMTProofsFreezings`applyCurrentFreezings[expr[[1]],freezingRenames],
				options
			];
			branchConditions={conditionExpression};
			collectStrongRenamings=False;
			collectWeakRenamings=False
	];
	
	FMTProofsBranches`beginBranches[
		rememberedLastCumulatedStatements,
		rememberedCumulatedAggregateAccesses,
		rememberedFreezingRenames,
		branchCumulatedStatements,
		branchCumulatedAggregateAccesses,
		conditionalFreezingUpdates,
		lastCumulatedStatements,
		cumulatedAggregateAccesses,
		freezingRenames
	];
	
	Scan[
		Function[branchPos,
			With[
				{
					branch=branchesToProcess[[branchPos]],
					branchCond=branchConditions[[branchPos]],
					otherBranchConditions=Delete[branchConditions,branchPos]
				},
				
				FMTProofsBranches`beginBranch[
					rememberedLastCumulatedStatements,
					rememberedCumulatedAggregateAccesses,
					rememberedFreezingRenames,
					lastCumulatedStatements,
					cumulatedAggregateAccesses,
					freezingRenames
				];
				
				(* recursive call for each identified branch: *)
				iterateNumbered[
					sys,
					packageName,
					qOperationName,
					opBody,
					allPathsStatements,
					lastCumulatedStatements,
					
					(* pre- and post-conditions have to be true
						only within the context of the branch
						and nested branches are Anded *)
					And[branchContext,branchCond],
					
					(* knowns are implied by the branch condition *)
					Function[st,
						cumulationWrapperFunction[
							Implies[branchCond,st]
						]
					],
					
					freezingRenames,
					lastFreezingIds,
					variableDomains,
					cumulatedAggregateAccesses,
					branch
				];
				
				FMTProofsBranches`endBranch[
					branchCond,
					otherBranchConditions,
					rememberedLastCumulatedStatements,
					rememberedCumulatedAggregateAccesses,
					rememberedFreezingRenames,
					branchCumulatedStatements,
					branchCumulatedAggregateAccesses,
					conditionalFreezingUpdates,
					lastCumulatedStatements,
					cumulatedAggregateAccesses,
					freezingRenames,
					collectStrongRenamings,
					True (* always collect weak renamings (those that bypass the branch) *)
				]
			]
		],
		
		Range[Length[branchesToProcess]]
	];
	
	(* conditionalFreezingUpdates has the form:
		
		<| condition \[Rule] <|name\[Rule]freezeRenaming, ...|>, ... |>
		
		The meaning of this structure is that when condition is True,
		the symbol name got new replacement; all branches need to be taken into account.
		The renamings, wrapped in implications based on branch conditions,
		are added to allPathsStatements, so that
		further statements can rely on this in their own proofs.
		The following call to endBranches does this.
	*)
	
	conditionalRenamingAssignments=FMTProofsBranches`endBranches[
		rememberedLastCumulatedStatements,
		rememberedCumulatedAggregateAccesses,
		rememberedFreezingRenames,
		branchCumulatedStatements,
		branchCumulatedAggregateAccesses,
		conditionalFreezingUpdates,
		lastCumulatedStatements,
		cumulatedAggregateAccesses,
		freezingRenames
	];
	
	(* process additional assignments that result from alternative renamings in branches: *)
	
	localVariablesThatNeedInitialization=
		First/@FMTSystem`operationUninitializedLocals[sys,qOperationName];
	Scan[
		Function[cond,
			condAssignments=conditionalRenamingAssignments[cond];
			Scan[
				Function[condAssignment,
					objectToRename=condAssignment[[1]];
					sourceObject=condAssignment[[2]];
					
					(* processing steps are as above for regular expressions: *)
					
					frozenKnowns=FMTProofsFreezings`freezeAndReplaceSingleAssignment[
						condAssignment,
						freezingRenames,
						lastFreezingIds,
						variableDomains,
						options,
						
						(* create new freeze for lhs if it was not yet done,
							but reuse the already created freeze for later assignments
							involving the same lhs object: *)
						makeNewLhsFreeze->FreeQ[objectsBranchRenamed,objectToRename],
						
						(* rhs in this assignment already properly
							indicates the right frozen source value,
							so there is no need to apply the freezings again: *)
						applyRhsFreezings->False
					];
					
					AppendTo[objectsBranchRenamed,objectToRename];
					
					aggregateKnowns=FMTProofsAggregates`processAggregateAccesses[
						sys,
						frozenKnowns,
						cumulatedAggregateAccesses
					];
					aggregateAccessPostKnownAssignments=aggregateKnowns[[1]];
					aggregateAccessPostKnownPropagations=aggregateKnowns[[2]];
					
					aggregateAccessPostKnownAssignments=
						FMTProofsFreezings`unwrapFrozenAndGrammar[
							aggregateAccessPostKnownAssignments,
							options
						];
					aggregateAccessPostKnownPropagations=
						FMTProofsFreezings`unwrapFrozenAndGrammar[
							aggregateAccessPostKnownPropagations,
							options
						];
					
					aggregateAccessPostKnownAssignments=
						DeleteCases[aggregateAccessPostKnownAssignments,True];
					
					aggregateAccessPostKnownPropagations=
						DeleteCases[aggregateAccessPostKnownPropagations,True];
					
					appendWrappedKnowns[
						aggregateAccessPostKnownAssignments,
						lastCumulatedStatements,
						Function[st,
							cumulationWrapperFunction[Implies[cond,st]]
						]
					];
					
					(* add also the conditional propagation of initialization flag,
						but only for those local variables that are initially uninitialized: *)
					If[MemberQ[localVariablesThatNeedInitialization,objectToRename[[1]]],
						AppendTo[
							lastCumulatedStatements,
							cumulationWrapperFunction[
								Implies[
									cond,
									Equal@@
										FMTSymbols`initializedT/@
											FMTProofsFreezings`unwrapFrozenAndGrammar[frozenKnowns[[1]],options]
								]
							]
						]
					]
				],
				condAssignments
			]
		],
		Keys[conditionalRenamingAssignments]
	]
]

SetAttributes[processIfWhileSwitch,HoldAll]

(* Helper for recursive calls, processes a single numberedT expression:
	get fact and wish statements related to that part
	and accumulates them on the path from the beginning of the operation's body
	(the statements are pre-processed by the cumulation wrapper function);
	accumulates all such statement lists with index numbers as well
*)
processSingleNumbered[
	sys_,
	packageName_,
	qOperationName_,
	opBody_,
	allPathsStatements_,
	lastCumulatedStatements_,
	branchContext_,
	cumulationWrapperFunction_,
	freezingRenames_,
	lastFreezingIds_,
	variableDomains_,
	cumulatedAggregateAccesses_,
	index_,
	expr_
]:=Module[
	{
		statementsForThisPart, (* both wishes and knowns for the given points *)
		preCondsForThisPart, (* preconditions *)
		postCondsForThisPart, (* postconditions *)
		preCallFreezings={}, (* freezing from before the operation call (used only with calls) *)
		knownsForThisPart, (* new statements resulting from object assignments *)		
		frozenKnowns={}, (* additional facts related to freezing of assigned variables,
							these should be further processes by aggregate access analysis *)

		aggregateKnowns, (* temp for aggregate analysis *)
		aggregateAccessPreKnownPropagations, (* separate knowns related to array indexing and struct field accesses *)
		aggregateAccessPostKnownAssignments, (* note: there are no aggregateAccessPreKnownAssignments here, *)
		aggregateAccessPostKnownPropagations, (* because PREconditions do not generate assignments,
												and POST values include also knowns from the operation itself *)
		options=FMTSystem`fmtGetCurrentOptions[sys]
	},
	
	(* get verification statements for this index: *)
	statementsForThisPart=FMTVerification`operationConditions[sys,packageName,qOperationName,opBody,index];
	
	(* statements are {preconds,postconds,knowns}: *)
	preCondsForThisPart=statementsForThisPart[[1]];
	postCondsForThisPart=statementsForThisPart[[2]];
	knownsForThisPart=statementsForThisPart[[3]];
	
	(* pre-conditions are resolved using the old freezing renames
		(before assigned variables get new symbol name): *)
	preCondsForThisPart=FMTProofsFreezings`applyCurrentFreezings[preCondsForThisPart,freezingRenames];
	
	(* freeze the symbol that gets new value in assignment: *)
	If[MatchQ[expr,_FMTSymbols`assignmentT],
		frozenKnowns=FMTProofsFreezings`freezeAndReplaceSingleAssignment[
			expr,
			freezingRenames,
			lastFreezingIds,
			variableDomains,
			options
		]
	];
	
	(* create new freezings for symbols that are used
		as inOut or out actual parameters in operation calls: *)
	If[MatchQ[expr,_FMTSymbols`callT],
		preCallFreezings=freezingRenames;
		
		(* expr is callT[qOpName,formalParamsT[...],actualParamsT[...]] *)
		(* here ScanThread was intended, but it does not exist: *)
		MapThread[
			Function[{formal,actual},
				With[
					{
						mode=Head[formal]
					},
					
					If[mode===FMTSymbols`inOut||mode===FMTSymbols`out,
						(* Note: no new knowns are created here, as was the case above
							with assignments; the reason for this is that assignments are
							direct sources of knowns (values for newly frozen symbols),
							whereas in operation calls the only knowns about the output
							parameters come from operation's post-conditions, which
							are dealt with elsewhere (in FMTVerification`) *)
						
						FMTProofsFreezings`freezeSingleSymbol[
							actual,freezingRenames,lastFreezingIds,variableDomains,options]
					]
				]
			],
			{List@@expr[[2]],List@@expr[[3]]}
		]
	];
	
	(* post-conditions and knowns should refer to new symbol name,
		if it was frozen in assignment: *)
	postCondsForThisPart=FMTProofsFreezings`applyCurrentFreezings[postCondsForThisPart,freezingRenames];
	knownsForThisPart=FMTProofsFreezings`applyCurrentFreezings[knownsForThisPart,freezingRenames];
	
	(* separate handling of aggregate accesses,
		operate only on those accesses that were used in collected proof statements;
		Note: the knowns related to aggregate accesses are divided into two groups:
			- known assignments, which result from aggregate assignments (these knowns
				reflect the propagation of values for individual fields and array elements),
				these are stored in lastCumulatedStatements and can be used in proofs
				generated immediately or for later points in the operation body,
			- known propagations, which are relations between fields and elements within
				the same aggregate object and involve array indexing,
				these are not stored permanently and are recreated every time,
				as their set will depend on the cumulated set of access paths that
				were used within the given operation
				(that is, instead of storing these knowns in lastCumulatedStatements,
				the cumulatedAggregateAccesses is a source to recreate them).
	*)
	aggregateKnowns=FMTProofsAggregates`processAggregateAccesses[
		sys,
		preCondsForThisPart[[All,2]],
		cumulatedAggregateAccesses
	];
	(* note: preconditions do not generate new assignments,
		only aggregate accesses that might require new value propagation knowns,
		-> there are no aggregateAccessPreKnownAssignments,
		and aggregateKnowns[[1]] is supposed to be empty *)
	Assert[aggregateKnowns[[1]]=={}];
	aggregateAccessPreKnownPropagations=aggregateKnowns[[2]];
	
	aggregateKnowns=FMTProofsAggregates`processAggregateAccesses[
		sys,
		Join[
			frozenKnowns,
			knownsForThisPart,
			postCondsForThisPart[[All,2]]
		],
		cumulatedAggregateAccesses
	];
	aggregateAccessPostKnownAssignments=aggregateKnowns[[1]];
	aggregateAccessPostKnownPropagations=aggregateKnowns[[2]];
	
	(* special handling of oldT[expr] - revert symbol replacements within this wrapper,
		note: this has to be done before unwrapping grammar, as this process relies
		on all wrappers to be in place *)
	preCondsForThisPart=restoreOldValues[preCondsForThisPart,preCallFreezings];
	knownsForThisPart=restoreOldValues[knownsForThisPart,preCallFreezings];
	postCondsForThisPart=restoreOldValues[postCondsForThisPart,preCallFreezings];
	
	preCondsForThisPart=FMTProofsFreezings`unwrapFrozenAndGrammar[preCondsForThisPart,options];
	postCondsForThisPart=FMTProofsFreezings`unwrapFrozenAndGrammar[postCondsForThisPart,options];
	knownsForThisPart=FMTProofsFreezings`unwrapFrozenAndGrammar[knownsForThisPart,options];
	aggregateAccessPreKnownPropagations=FMTProofsFreezings`unwrapFrozenAndGrammar[aggregateAccessPreKnownPropagations,options];
	aggregateAccessPostKnownAssignments=FMTProofsFreezings`unwrapFrozenAndGrammar[aggregateAccessPostKnownAssignments,options];
	aggregateAccessPostKnownPropagations=FMTProofsFreezings`unwrapFrozenAndGrammar[aggregateAccessPostKnownPropagations,options];
	
	(* sanitize preconditions and knowns by removing tautologies and duplicates: *)
	preCondsForThisPart=DeleteDuplicates[
		DeleteCases[preCondsForThisPart,
			{_String,FMTSymbols`alwaysT[True]}|{_String,FMTSymbols`neverT[False]}
		]
	];
	aggregateAccessPreKnownPropagations=DeleteCases[aggregateAccessPreKnownPropagations,True];
	aggregateAccessPostKnownAssignments=DeleteCases[aggregateAccessPostKnownAssignments,True];
	aggregateAccessPostKnownPropagations=DeleteCases[aggregateAccessPostKnownPropagations,True];
	
	(* Note: preconditions are considered *without* new knowns that are contributed
		at this point - the reason for this is that new knowns might divert the solver
		from finding examples (or counter-examples) for the precondition,
		as they can unnecessarily constrain the solution domain, for example:
		
		y=1/x;
		
		The precondition for the above is x\[NotEqual]0 and the new known
		(with frozen symbol for y) is y$n=1/x. If the domain for both {x,y} is Integers,
		then the counterexample for x\[Equal]0 will not be found, as the new known implicitly
		constrains the solution domain to legal values of x.
		
		The solution to this problem is to gather preconditions without new knowns
		and then, separately, postconditions with new knowns. This also means
		that in the complete result there can be 1 or 2 entries for the given index.
	*)
	
	(* 1. preconditions without new knowns,
		but with current branch context, if non-trivial: *)
	
	If[preCondsForThisPart!={},
		AppendTo[allPathsStatements,
			{
				index,
				Join[
					lastCumulatedStatements,
					aggregateAccessPreKnownPropagations,
					If[branchContext=!=True,List[branchContext],{}]
				],
				preCondsForThisPart
			}
		]
	];
	
	(* 2. accumulate new knowns: *)
	
	appendWrappedKnowns[knownsForThisPart,lastCumulatedStatements,Identity];
	appendWrappedKnowns[aggregateAccessPostKnownAssignments,lastCumulatedStatements,Identity];
	appendWrappedKnowns[aggregateAccessPostKnownPropagations,lastCumulatedStatements,Identity];
	
	(* 3. postconditions take into account new knowns
		and the current branch context, if non-trivial: *)
	
	If[(postCondsForThisPart!={})||
		(knownsForThisPart!={})||
		(aggregateAccessPostKnownAssignments!={})||
		(aggregateAccessPostKnownPropagations!={}),
		
		AppendTo[allPathsStatements,
			{
				index,
				Join[
					lastCumulatedStatements,
					If[branchContext=!=True,List[branchContext],{}]
				],
				postCondsForThisPart
			}
		]
	];
	
	(* If the part is a nested numberedT (for example, If, While, Switch, etc.),
		then remember current value of lastCumulatedStatements
		and recurse on all available branches;
		revert to the remembered value after recursion
	*)
	If[MemberQ[{FMTSymbols`ifT,FMTSymbols`whileT,FMTSymbols`switchT},Head[expr]],
		processIfWhileSwitch[
			sys,
			packageName,
			qOperationName,
			opBody,
			allPathsStatements,
			lastCumulatedStatements,
			branchContext,
			cumulationWrapperFunction,
			freezingRenames,
			lastFreezingIds,
			variableDomains,
			cumulatedAggregateAccesses,
			expr
		]
	]
]

SetAttributes[processSingleNumbered,HoldAll]


(* Helper, prepares a list of paths with logic statements from the beginning of the operation
	to the index where the path ends
*)
iteratePathsAndGetFactStatements[
	sys_,
	packageName_,
	qOperationName_,
	opBody_,
	variableDomains_
]:=Module[
	{
		freezingRenames=<||>,
		lastFreezingIds=<||>,
		lastCumulatedStatements={}, (* {statement,...} *)
		cumulatedAggregateAccesses=<||>,
		allPathsStatements={} (* {{index,{statement,...}},...} *)
	},
	
	(* Note: allPathsStatements is updated by calls to processSingleNumbered *)
	
	(* 1. gather statements for the local "start" point
		(for example: variables are not initialized at the start): *)
	processSingleNumbered[
		sys,
		packageName,
		qOperationName,
		opBody,
		allPathsStatements,
		lastCumulatedStatements,
		True, (* dummy branch context at the operation's top-level *)
		Identity, (* dummy cumulation wrapper at the operation's top-level *)
		freezingRenames,
		lastFreezingIds,
		variableDomains,
		cumulatedAggregateAccesses,
		"start",
		Null (* dummy *)
	];
	
	(* 2. process recursively all numberedT in the operation's body: *)
	iterateNumbered[
		sys,
		packageName,
		qOperationName,
		opBody,
		allPathsStatements,
		lastCumulatedStatements,
		True, (* dummy branch context at the operation's top-level *)
		Identity, (* dummy cumulation wrapper at the operation's top-level *)
		freezingRenames,
		lastFreezingIds,
		variableDomains,
		cumulatedAggregateAccesses,
		opBody (* start with the whole operation's body (iterations are done at level {1}),
					further recursions (on conditional branches and loops) will narrow this scope *)
	];
	
	(* 3. finally, accumulate operation post-conditions: *)
	processSingleNumbered[
		sys,
		packageName,
		qOperationName,
		opBody,
		allPathsStatements,
		lastCumulatedStatements,
		True, (* dummy branch context at the operation's top-level *)
		Identity, (* dummy cumulation wrapper at the operation's top-level *)
		freezingRenames,
		lastFreezingIds,
		variableDomains,
		cumulatedAggregateAccesses,
		"finish",
		Null (* dummy *)
	];
	
	allPathsStatements
]

SetAttributes[iteratePathsAndGetFactStatements,HoldAll]


(* Returns the list of proof statements in the form:
	{
		{
			index,
			knowns,
			{{description,wish},...}
		},
		...
	},
	
	where index identifies the place in the operation where the proof statement was generated
	(this can be used to highlight that place in grammar or control flow diagrams),
	where knowns is a list of frozen proof statements until (and excluding) the "wish" point,
	and wishes is a list of {description,wish} pairs,
	where wish is a statement that needs to be true in order for the operation to be correct.
	
	In the simplest case, the whole list might be empty - for example when the operation itself
	has no postconditions (that is, nothing to wish for at the end of operation) and no precondition
	in any of its internal statements.
	If the operation has postconditions, these are wishes that accompany all knowns from the whole
	operation.
	Note: wishes have symbols replaced with freezingRenames that were active at the wish point.
	
	In addition, variableDomains is an Association that is filled with domain names
	for all variables used in the proof statements.
*)
proofStatements[sys_,packageName_,qOperationName_,variableDomains_]:=Module[
	{
		opBody, (* resolved operation body *)
		typeHier,
		domains, (* temporary, to find variable domains as used in the operation body and proof statements *)
		allPathStatements={} (* {{index,facts,{{description,wish},...},reverseFreezes},...} - list of known facts
								from the beginning of the operation to the given point,
								together with their verification conditions (wishes);
								indices include "finish" *)
	},
	
	opBody=FMTSystem`resolvedOperationBody[sys,packageName,qOperationName,True,{}];
	
	(* opBody can be Null if the given operation is declared, but not defined *)
	If[opBody=!=Null,
		typeHier=FMTSystem`typeHierarchy[sys];
		
		(* extract domains for those objects that are used in the operation: *)
		domains=FMTSystem`operationAllMentionedObjects[sys,packageName,qOperationName,opBody];
		
		(* turn program-level types into proof domain names: *)
		
		(* 1. user types -> standard types *)
		domains=domains/.typeHier;
		
		(* 2. standard types \[Rule] domains *)
		domains=domains/.{FMTSymbols`booleanT->Booleans,FMTSymbols`integerT->Integers};
		
		AppendTo[variableDomains,domains];
		
		allPathStatements=iteratePathsAndGetFactStatements[
			sys,
			packageName,
			qOperationName,
			opBody,
			variableDomains
		]
	];
	
	allPathStatements
]

SetAttributes[proofStatements,HoldAll]


End[];


EndPackage[];
