(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


Needs["FMTProofsFreezings`"];
Needs["FMTSymbols`"];
Needs["FMTSystem`"];


BeginPackage["FMTProofsAggregates`"];


processAggregateAccesses::usage="processAggregateAccesses[sys,expression,cumulatedAggregateAccesses] Maintains the cumulatedAggregateAccesses data structure.";


Begin["Private`"];


(* Helpers, identify the top-level object in the aggregate access
	by removing the outermost (which from the code perspective is the leaf of the branch)
	partT or dotT segment
*)
accessDropStep[object_]:=Module[
	{
		innerHead  (* Head of what is directly inside the object *)
	},
	
	innerHead=Head[object[[1]]];
	Switch[innerHead,
		FMTSymbols`partT|FMTSymbols`dotT,
			object[[1,1]],
		
		FMTProofsFreezings`frozenT|FMTSymbols`objectT|FMTSymbols`objectParentT|Symbol,
			object
	]
]


(* Identifies the name of top-level object in the aggregate access *)
accessTopObject[object_]:=FixedPoint[accessDropStep,object]


(* Helper, identifies a single step of aggregate access (partT or dotT).
	Due to how grammar is represented, the most outer access operation
	referes to the leaf of the branch that is visible within the whole data structure tree.
	For example, with s.a[[i]].f the dotT[f] is peeled and appended to path,
	and the function returns the pruned branch s.a[[i]].
	If there is nothing to peel, the function returns obj.
*)
accessPeelStep[obj_,path_]:=Module[
	{
		innerHead , (* Head of what is directly inside the object *)
		fieldOrIdx, (* name of field within the structure or array index *)
		type (* type of referenced element, whether array element or structure field *)
	},
	
	If[Head[obj]=!=Symbol,
		innerHead=Head[obj[[1]]];
		Switch[innerHead,
			FMTSymbols`partT|FMTSymbols`dotT,
				fieldOrIdx=obj[[1,2]];
				type=obj[[2]];
				AppendTo[path,innerHead[fieldOrIdx,type]];
				obj[[1,1]],
			
			FMTSymbols`objectT|FMTSymbols`objectParentT|FMTSymbols`objectLhsT,
				
				ReplacePart[obj,{1,0}->FMTSymbols`objectT],
			
			Symbol,
				obj
		],
		obj
	]
]

SetAttributes[accessPeelStep,HoldRest]


(* Extracts the complete access path from the subobject reference
	and returns the rule (appropriate for merging into an Association object):
	{obj,ver}\[Rule]{path}
	where path can be an empty list for simple objects.
*)
accessPath[object_]:=Module[
	{
		topObj,
		obj,
		freezeVer,
		path={}
	},
	
	topObj=FixedPoint[Function[obj,accessPeelStep[obj,path]],object];
	Switch[Head[topObj],
		FMTProofsFreezings`frozenT,
			obj=topObj[[1]];
			freezeVer=topObj[[2]],
		
		FMTSymbols`objectT|FMTSymbols`objectParentT|FMTSymbols`objectLhsT,
			obj=FMTSymbols`objectT@@topObj;
			freezeVer=0,
		
		Symbol,
			obj=topObj;
			freezeVer=0
	];
	
	{obj,freezeVer}->{Reverse[path]}
]


(* Applies a single access step to the given object;
	the step might be an array indexing or a structure field reference *)
applyPathStep[obj_,pathSegment_]:=Module[
	{
		innerHead , (* Head of the wrapping operation (partT or dotT) *)
		fieldOrIdx, (* name of field within the structure or array index *)
		type, (* type of referenced element, whether array element or structure field *)
		objectWrapperSanitation
	},
	
	(* the object wrapper should be changed to objectParentT for partT and dotT accesses: *)
	objectWrapperSanitation={
		FMTSymbols`objectT->FMTSymbols`objectParentT,
		FMTSymbols`objectLhsT->FMTSymbols`objectParentT
	};
	
	innerHead=Head[pathSegment];
	Switch[innerHead,
		FMTSymbols`partT|FMTSymbols`dotT,
			fieldOrIdx=pathSegment[[1]];
			type=pathSegment[[2]];
			FMTSymbols`objectT[innerHead[obj/.objectWrapperSanitation,fieldOrIdx],type],
		
		_,
			Print["TODO: unrecognized access path segment: ",pathSegment];
	]
]

(* Applies the whole access path to the given object
	(which itself might be already a subobject of something bigger),
	if the given object is a top-level object, the access path is an empty list. *)
applyAccessPath[obj_,path_]:=Fold[applyPathStep,obj,path]


(* Returns {True,postfixPath} if the access path of object is equal or is an initial subsequence of path,
	with field segments matching exactly, and array index segments allowed to have different indices;
	If the match is found, postfixPath is the remaing part of path (which can be an empty list, if the whole of path matches)
	otherwise {False,{}} is returned, meaning that the access paths are different and do not refer to the same subobject.
*)
fieldPathPrefixMatch[path_,object_]:=Module[
	{
		objectPath,
		objectPathLen,
		postfixPath={}
	},
	
	objectPath=accessPath[object][[2,1]]; (* accessPath gives {obj,ver}\[Rule]{path} *)
	objectPathLen=Length[objectPath];
	
	If[(objectPathLen<=Length[path])&&
		And@@MapThread[
			Function[{objectPathStep,pathStep},
				With[
					{
						pathStepInnerHead=Head[pathStep],
						objectPathStepInnerHead=Head[objectPathStep]
					},
					
					(pathStepInnerHead===objectPathStepInnerHead)&&
						((pathStepInnerHead===FMTSymbols`dotT)\[Implies](pathStep===objectPathStep))
				]
			],
			{objectPath,Take[path,objectPathLen]}
		],
		
		{True,Drop[path,objectPathLen]},
		{False,{}}
	]
]


(* Checks whether the given path finishes on access to some simple data type *)
simpleTypePathQ[sys_,path_]:=
	FMTSystem`simpleTypeQ[sys,Last[path][[2]]]


(* Helper, returns True if the given object is of a simple type,
	subObject is objectT[...] or frozenT[objectT[...]] or similar structure *)
simpleTypeSubObjectQ[sys_,subObject_]:=Module[
	{
		type
	},
	
	If[Head[subObject]===FMTProofsFreezings`frozenT,
		type=subObject[[1,2]],
		type=subObject[[2]]
	];
	
	FMTSystem`simpleTypeQ[sys,type]
]


(* Create the condition for equal values of all partT indices that are used
	on the same positions in both path1 and path2;
	note: does not assume structural similarity of paths,
	if paths have different structure (not matching segments), then the condition is False
*)
equalIndicesExpression[path1_,path2_]:=Module[
	{
		conditions={}
	},
	
	If[Length[path1]==Length[path2],
		conditions=MapThread[
			Function[{segment1,segment2},				
				SameQ[segment1,segment2]||
					(SameQ[Head[segment1],Head[segment2],FMTSymbols`partT]&&
						segment1[[1]]==segment2[[1]] (* equal array indices *)
					)
			],
			{path1,path2}
		],
		
		conditions={False}
	];
	
	And@@conditions
]


(* Creates the condition for different values of all partT indices that are used
	on the same positions in both path1 and path2;
	note: assumes that the path structure
	was already verified (that is, partT segments are indeed on the same positions)
*)
unequalIndicesExpression[path1_,path2_]:=With[
	{
		indices1=Cases[path1,FMTSymbols`partT[idx_,_]->idx,{1}],
		indices2=Cases[path2,FMTSymbols`partT[idx_,_]->idx,{1}]
	},
	
	Or@@Thread[Unequal[indices1,indices2]]
]


(* Helper, extracts object accesses from the given expression *)
extractObjects[expression_]:=Module[
	{
		work,
		temp,
		result={}
	},
	
	(* take into account also subexpressions used in array indexing operations
		- they also can contain object accesses, like in a1[[a2[[3]]]]
	*)
	work=Flatten[List[expression,Cases[expression,FMTSymbols`partT[_,idx_]->idx,All]]];
	
	While[work!={},
		temp={};
		Scan[
			Function[ex,
				Switch[Head[ex],
					FMTSymbols`objectLhsT|FMTSymbols`objectT|FMTProofsFreezings`frozenT,
						AppendTo[result,ex],
					FMTSymbols`initializedT|Symbol|Integer|Rational,
						(* abandon these elements,
							they do not contribute to the list of accessed objects *)
						Null,
					_,
						(* change into List to process its elements in next turn *)
						AppendTo[temp,List@@ex]
				]
			],
			work
		];
		work=Flatten[temp]
	];
	
	DeleteDuplicates[result]
]

(* Helper, returns the list of keys denoting newer (with higher freezing ids) objects
	with regard to the given one *)
newerObjects[cumulatedAggregateAccesses_,key_]:=
	Select[Keys[cumulatedAggregateAccesses],
		Function[k,
			k[[1]]==key[[1]]&&k[[2]]>key[[2]]
		]
	]

(* Helper, returns the list of paths accessing newer versions of the given object *)
newerPaths[cumulatedAggregateAccesses_,key_]:=With[
	{
		objects=newerObjects[cumulatedAggregateAccesses,key]
	},
	
	Flatten[Lookup[cumulatedAggregateAccesses,objects],1]
]

(* Helper, returns True if newerPaths contains a paths that is overwriting
	(that is, it is the same path or is its shorter prefix) the given path *)
isOverwrittenQ[path_,newerPaths_]:=
	Or@@Map[
		Function[p,
			Length[path]>Length[p]&&Take[path,Length[p]]==p
		],
		newerPaths
	]

(* Helper, removes those access paths that were overwritten by accesses in newer
	versions of the same objects *)
pruneOverwrittenAccesses[cumulatedAggregateAccesses_]:=Module[
	{
		key,
		paths,
		newerAccessPaths,
		retainedPaths,
		result=<||>
	},
	
	Scan[
		Function[access, (* {aggregate-name,ver} \[Rule] {freezing-access-path,access-path,...} *)
			key=access[[1]];
			paths=access[[2]];
			newerAccessPaths=newerPaths[cumulatedAggregateAccesses,key];
			retainedPaths=Select[paths,
				Function[p,
					Not[isOverwrittenQ[p,newerAccessPaths]]
				]
			];
			
			If[retainedPaths!={},
				result[key]=retainedPaths
			]
		],
		Normal[cumulatedAggregateAccesses]
	];
	
	result
]

(* Helper, merges new access paths into the main cumulatedAggregateAccesses data structure *)
manageAggregateAccesses[expression_,cumulatedAggregateAccesses_]:=Module[
	{
		aggregateObjects,
		aggregateAccesses
	},
	
	(* extract full object accesses from the given expression,
		Note: a[[idx]] is an example of complex access, which mentions
			the whole object a, but that should not be counted as
			accessing the whole object - only the actual indexing path
			should be reported. *)
	
	aggregateObjects=extractObjects[expression];
	
	aggregateAccesses=Flatten[
		Map[accessPath,aggregateObjects],
		{1}
	];
	
	(* Note: conveniently, merging of the rules above will pick the first
		occurence of the given version first, as the ordering of all entities
		is preserved in the statements list and in the operations above
		- that is, the first value in the combined list for the given key
			is the access path that was used to *modify* (and freeze)
			the given array or structure object, if its freezing index is positive
			(if the freezing index is 0, then it is the original object, not a new freeze,
			and all access paths in the list were reading accesses)
	*)
	If[aggregateAccesses!={},
		cumulatedAggregateAccesses=Merge[
			{cumulatedAggregateAccesses,aggregateAccesses},
			Function[lst,DeleteDuplicates[Flatten[lst,1]]]
		]
	]
]

SetAttributes[manageAggregateAccesses,HoldRest]


(* Helper *)
doKnownAssignments[sys_,expression_,cumulatedAggregateAccesses_]:=Module[
	{
		assignments,
		knownAssignments={}
	},
	
	assignments=Cases[expression,_FMTSymbols`assignmentT,All];
	
	Scan[
		Function[assig, (* assignmentT[lhs,rhs] *)
			Module[
				{
					type,
					lhs,
					lhsSubobject,
					rhs,
					rhsTopObject,rhsTopObjectName,
					rhsVer=0,
					rhsPath,
					rhsSubobjectNoVer, (* freeze id will be added as needed *)
					previousAccessPaths,
					pathMatchResult,
					pathMatches,
					postfixPath,
					previousSubobject,
					propagationStatement
				},
				
				lhs=assig[[1]];
				rhs=assig[[2]];
				
				If[MemberQ[FMTSymbols`simpleExpressionTags,Head[rhs]],
					(* assignment from literals or simple expressions do not propagate values from other objects,
						but directly from the right-hand side literal or expression: *)
					AppendTo[knownAssignments,Equal[lhs,rhs]],
					
					(* assignment from other object might require value propagation
						from all previously accessed subobjects of rhs: *)
					type=Cases[lhs,FMTSymbols`objectLhsT[_,t_]->t,All][[1]];
					rhsTopObject=accessTopObject[rhs];
					If[Head[rhsTopObject]===FMTProofsFreezings`frozenT,
						rhsVer=rhsTopObject[[2]];
						rhsTopObject=rhsTopObject[[1]]
					];
					
					rhsTopObjectName=rhsTopObject[[1]];
					rhsPath=accessPath[rhs];
					rhsSubobjectNoVer=applyAccessPath[rhsTopObject,rhsPath[[2,1]]];
					
					(* find access paths that were used so far for the rhs object *)
					previousAccessPaths=Lookup[cumulatedAggregateAccesses,Key[{rhsTopObject,rhsVer}],{}];
					
					Scan[
						Function[previousPath,
							pathMatchResult=fieldPathPrefixMatch[previousPath,rhsSubobjectNoVer];
							pathMatches=pathMatchResult[[1]];
							If[pathMatches,
								(* The rhs subobject is an initial subpath of one of the
									previous paths that were used to access the same aggregate object;
									propagate values from previous access, keeping the trailing
									(postfix) part of the previous access path, as it relates
									to the sub-parts of the newly modified lhs
								*)
								postfixPath=pathMatchResult[[2]];
								lhsSubobject=applyAccessPath[lhs,postfixPath];
								
								(* value propagation: *)
								If[simpleTypeSubObjectQ[sys,lhsSubobject],
									previousSubobject=applyAccessPath[rhsSubobjectNoVer,postfixPath];
									
									propagationStatement=Equal[
										lhsSubobject,
										FMTProofsFreezings`applySingleFreezing[
											previousSubobject,
											rhsTopObject,
											rhsVer
										]
									];
									
									AppendTo[knownAssignments,propagationStatement];
								];
								
								(* additionally, add the newly touched lhs subobject
									as an access path for lhs object, so that it can be picked
									by the analysis of further statements:
								*)
								appendNewAggregateAccesses[lhsSubobject,cumulatedAggregateAccesses]
							]
						],
						previousAccessPaths
					]
				]
			]
		],
		assignments
	];
	
	knownAssignments
]

SetAttributes[doKnownAssignments,HoldRest]


(* Helper *)
doKnownPropagationsAcrossHistory[sys_,newKeys_,cumulatedAggregateAccesses_]:=Module[
	{
		knownPropagations={}
	},
	
	Scan[
		Function[key, (* {aggregate-name,ver}, this is one of new keys in this structure *)
			Module[
				{
					frozenAggregateName,
					aggregateType,
					aggregateVer,
					accessPaths,
					freezingPath,
					previousAggregateVer,
					previousAccessPaths,
					pathMatchResult,pathMatches,postfixPath,
					previousSubobject,
					newlyFrozenSubobject
				},
				
				frozenAggregateName=key[[1]];
				aggregateVer=key[[2]];
				accessPaths=cumulatedAggregateAccesses[key];
				
				If[aggregateVer > 0,
					(* this is the aggregate that was already affected by assignment (frozen),
						the first access path in accessPaths was created for a new freeze,
						and new equivalence rules should be generated
						to indicate that the old values were not overwritten
						and are present also in the new aggregate version like here, for arrays:
						
						i \[NotEqual] j => Subscript[a, n][[j]] \[Equal] Subscript[a, n-1][[j]]
						
						for structure fields it is simpler, because no indices are involved;
						such rules are created by visiting all previous versions
						of the aggregate and picking the indices and fields that were used there,
						paired with the new indices/fields that were used to freeze the new version
					*)
					
					aggregateType=frozenAggregateName[[2]];
					freezingPath=accessPaths[[1]];
					
					previousAggregateVer=aggregateVer-1;
					
					(* Note: it is possible that ver.0 is not present in cumulatedAggregateAccesses,
						if the first use of the given object was to modify it -
						that is, the first use is freezing and ver.1 is immediately created,
						with no accesses recorded for the original object (ver.0)
						- assume implicit empty path for such missing versions:
					*)
					previousAccessPaths=Lookup[cumulatedAggregateAccesses,Key[{frozenAggregateName,previousAggregateVer}],{}];
					
					Scan[
						Function[previousPath,
							pathMatchResult=fieldPathPrefixMatch[previousPath,frozenAggregateName];
							pathMatches=pathMatchResult[[1]];
							If[pathMatches,
								(* The freezing path is an initial subpath of one of the
									previous paths that were used to access the same aggregate object;
									propagate values from previous access, keeping the trailing
									(postfix) part of the previous access path, as it relates
									to the sub-part of the newly created freeze
								*)
								postfixPath=pathMatchResult[[2]];
								
								previousSubobject=applyAccessPath[frozenAggregateName,previousPath];
								newlyFrozenSubobject=applyAccessPath[frozenAggregateName,postfixPath];
								
								(* value propagation:
									if the old value (the value within the previous freeze of the whole aggregate)
									was not overwritten, then it is kept in the new freeze at the same position
								*)
								
								AppendTo[knownPropagations,
									
									(* i \[NotEqual] j => Subscript[a, n][[j]] \[Equal] Subscript[a, n-1][[j]] *)
									(* Note: if the access path contains multiple partT segments,
										then the left side of this implication is:
										i \[NotEqual] p \[And] j \[NotEqual] q \[And] k \[NotEqual] r \[And] ...
										where {i,j,k,...} and {p,q,r,...} are indices
										on subsequent partT segments in both paths;
										similarly, if there are no partT segment whatsoever,
										then the left side is not needed (it can be True, to preserve expression structure).
									*)
									Implies[
										unequalIndicesExpression[freezingPath,previousPath],
										Equal[
											FMTProofsFreezings`applySingleFreezing[
												newlyFrozenSubobject,
												frozenAggregateName,
												aggregateVer
											],
											FMTProofsFreezings`applySingleFreezing[
												previousSubobject,
												frozenAggregateName,
												previousAggregateVer
											]
										]
									]
								]
							]
						],
						
						(* propagate only simple values: *)
						If[FMTSystem`simpleTypeQ[sys,aggregateType],
							(* if the aggreate is a simple type,
								previousAccessPaths is {} anyway,
								but the above code should deal with it properly: *)
							previousAccessPaths,
							
							(* if the aggregate is a complex type,
								pick only those paths that refer to simple (leaf)
								components of the aggregate: *)
							Select[previousAccessPaths,
								Function[path,
									path!={}&&simpleTypePathQ[sys,path]
								]
							]
						]
					]
				]
			]
		],
		newKeys
	];
	
	knownPropagations
]

SetAttributes[doKnownPropagationsAcrossHistory,HoldRest]


(* Helper *)
doKnownPropagationsWithinAggregate[newKeys_,cumulatedAggregateAccesses_]:=Module[
	{
		knownPropagations={}
	},
	
	Scan[
		Function[key, (* {aggregate-name,ver}, one of new keys in this structure *)
			Module[
				{
					aggregateName,
					aggregateVer,
					accessPaths,
					pathPairs,
					path1,
					path2
				},
				
				aggregateName=key[[1]];
				aggregateVer=key[[2]];
				accessPaths=cumulatedAggregateAccesses[key];
				
				pathPairs=Subsets[accessPaths,{2}];
				Scan[
					Function[pp, (* {path1,path2} *)
						path1=pp[[1]];
						path2=pp[[2]];
						
						AppendTo[knownPropagations,
							Implies[
								equalIndicesExpression[path1,path2],
								Equal[
									FMTProofsFreezings`applySingleFreezing[
										applyAccessPath[aggregateName,path1],
										aggregateName,
										aggregateVer
									],
									FMTProofsFreezings`applySingleFreezing[
										applyAccessPath[aggregateName,path2],
										aggregateName,
										aggregateVer
									]
								]
							]
						]
					],
					pathPairs
				]
			]
		],
		newKeys
	];
	
	knownPropagations
]

SetAttributes[doKnownPropagationsWithinAggregate,HoldRest]


(* Maintains the cumulatedAggregateAccesses data structure,
	which is an Association, where
	- keys are {name,ver} pairs, where name is the original name of the aggregate object
		and ver is its freezing version number (0 meaning the original object),
	- values are lists of access paths used on the given object: {path1,path2,...},
		where each path is itself a list of access segments partT[idx] or dotT[fieldName],
	so for example {a,1}\[Rule]{{partT[literalT[5,integerT]]},{partT[literalT[6,integerT]]}}
	means that in the operation body analyzed so far, there were two accesses
	to the first freeze of object a:
		Subscript[a,1][[5]] and Subscript[a,1][[6]]
	By convention, for entries with ver>0 the first path on the list
		was a modifying (freezing) access.
	This function returns {knownAssignments,knownPropagations}, where
	- knownAssignments are statements resulting from assignments between aggregate objects
		(and these should be added permanently
		to the list of cumulated statements for the given operation),
	- knownPropagations are statements about other subobject values being equal,
		(and these are always recreated from scratch, so should not be stored)
*)
processAggregateAccesses[sys_,expression_,cumulatedAggregateAccesses_]:=Module[
	{
		aggregateAssignments,
		knownAssignments={},
		knownPropagationsAcrossHistory={},
		knownPropagationsWithinAggregate={},
		tempCumulatedAggregateAccesses,
		keys,newKeys
	},
	
	keys=Keys[cumulatedAggregateAccesses];
	
	manageAggregateAccesses[expression,cumulatedAggregateAccesses];
	
	newKeys=Complement[Keys[cumulatedAggregateAccesses],keys];
	
	(* 1. generate knowns related to value propagation between aggregate objects
		(based on assignment statements only, but if the assignment is between whole
		objects or whole subobjects, the value propagation is generated for all
		known access paths that possibly affect the result from rhs to lhs): *)
	
	knownAssignments=doKnownAssignments[sys,expression,cumulatedAggregateAccesses];
	
	(* 1a. take the newly generated access paths into account in the access path database,
		and prune obsolete access paths, to avoid further steps generating duplicate assignments: *)
	
	manageAggregateAccesses[knownAssignments,cumulatedAggregateAccesses];
	
	tempCumulatedAggregateAccesses=cumulatedAggregateAccesses;
	If[knownAssignments!={},
		tempCumulatedAggregateAccesses=pruneOverwrittenAccesses[tempCumulatedAggregateAccesses]
	];
	
	(* 2. generate knowns related to value propagation across the history of a single data structures,
		both for arrays and for structure fields
		(that is, if the new version of lhs is not entirely overwritten,
		some parts of its previous version will be propagated to the new version of lhs): *)
	
	knownPropagationsAcrossHistory=doKnownPropagationsAcrossHistory[sys,newKeys,tempCumulatedAggregateAccesses];
	
	(* 3. generate knowns related to value equivalents within the same aggregate version
		if there are multiple accesses with equal array indices, that is:
		
		i \[Equal] j => a[[i]] \[Equal] a[[j]]
		
		(in other words, if the new version of lhs is accessed by several paths
		that might be possibly equivalent, like i and j indices above,
		then the equivalence of accessed elements is also stated)
	*)
	
	knownPropagationsWithinAggregate=doKnownPropagationsWithinAggregate[newKeys,tempCumulatedAggregateAccesses];
	
	(* 4. take the newly generated access paths into account in the access path database,
		and prune obsolete access paths: *)
	
	manageAggregateAccesses[
		{knownPropagationsAcrossHistory,knownPropagationsWithinAggregate},
		cumulatedAggregateAccesses
	];
	
	{knownAssignments,Join[knownPropagationsAcrossHistory,knownPropagationsWithinAggregate]}
]

SetAttributes[processAggregateAccesses,HoldRest]


End[];


EndPackage[];
