(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


Needs["FMTSymbols`"];


BeginPackage["FMTProofsBranches`"];


beginBranches::usage="beginBranches[...] Helper for managing conditional state within branches.";


endBranches::usage="endBranches[...] Helper for managing conditional state within branches,"<>
"returns the association structure with the list of assignment for conditional renamings.";


beginBranch::usage="beginBranch[...] Helper for managing conditional state within branches.";


endBranch::usage="endBranch[...] Helper for managing conditional state within branches.";


Begin["Private`"];


(* Helper, creates renaming assignment for those symbols
	that got new freezing renamings in conditional branches;
	conditionalFreezingUpdates is <| condition\[Rule]<| name\[Rule]freezeRenaming, ... |>, ... |>
*)
makeConditionalRenaming[conditionalFreezingUpdates_]:=Module[
	{
		assignment,
		freezes,
		sourceValue,
		result=<||>
	},
	
	Scan[
		Function[cond,
			freezes=conditionalFreezingUpdates[cond];
			
			result[cond]=Map[
				Function[name,
					sourceValue=freezes[name];
					
					(* synthesised assignment statement: *)
					FMTSymbols`assignmentT[FMTSymbols`objectLhsT@@name,sourceValue]
				],
				Keys[freezes]
			]
		],
		Keys[conditionalFreezingUpdates]
	];
	
	result
]


(* Helper, reduces the association by leaving only objectT keys *)
onlyObjectTKeyedEntries[assoc_]:=
	Association@@Rule@@@Cases[List@@@Normal[assoc],{_FMTSymbols`objectT,_},All]


(* Helpers for managing branches *)
beginBranches[
	rememberedLastCumulatedStatements_,
	rememberedCumulatedAggregateAccesses_,
	rememberedFreezingRenames_,
	branchCumulatedStatements_,
	branchCumulatedAggregateAccesses_,
	conditionalFreezingUpdates_,
	lastCumulatedStatements_,
	cumulatedAggregateAccesses_,
	freezingRenames_
]:=(
	rememberedLastCumulatedStatements=lastCumulatedStatements;
	rememberedCumulatedAggregateAccesses=cumulatedAggregateAccesses;
	rememberedFreezingRenames=freezingRenames;
	branchCumulatedStatements={};
	branchCumulatedAggregateAccesses=<||>;
	conditionalFreezingUpdates=<||>
)

SetAttributes[beginBranches,HoldAll]


endBranches[
	rememberedLastCumulatedStatements_,
	rememberedCumulatedAggregateAccesses_,
	rememberedFreezingRenames_,
	branchCumulatedStatements_,
	branchCumulatedAggregateAccesses_,
	conditionalFreezingUpdates_,
	lastCumulatedStatements_,
	cumulatedAggregateAccesses_,
	freezingRenames_
]:=(
	lastCumulatedStatements=Join[rememberedLastCumulatedStatements,
		branchCumulatedStatements
	];
	
	cumulatedAggregateAccesses=Join[rememberedCumulatedAggregateAccesses,
		branchCumulatedAggregateAccesses
	];
	
	freezingRenames=rememberedFreezingRenames;
	
	makeConditionalRenaming[conditionalFreezingUpdates]
)

SetAttributes[endBranches,HoldAll]


beginBranch[
	rememberedLastCumulatedStatements_,
	rememberedCumulatedAggregateAccesses_,
	rememberedFreezingRenames_,
	lastCumulatedStatements_,
	cumulatedAggregateAccesses_,
	freezingRenames_
]:=(
	lastCumulatedStatements=rememberedLastCumulatedStatements;
	cumulatedAggregateAccesses=rememberedCumulatedAggregateAccesses;
	freezingRenames=rememberedFreezingRenames
)

SetAttributes[beginBranch,HoldAll]


endBranch[
	branchCondition_,
	otherBranchConditions_,
	rememberedLastCumulatedStatements_,
	rememberedCumulatedAggregateAccesses_,
	rememberedFreezingRenames_,
	branchCumulatedStatements_,
	branchCumulatedAggregateAccesses_,
	conditionalFreezingUpdates_,
	lastCumulatedStatements_,
	cumulatedAggregateAccesses_,
	freezingRenames_,
	includeStrongRenamings_,
	includeWeakRenamings_
]:=Module[
	{
		newRenamesInThisBranch, (* Association *)
		newRenamedNames, (* list of symbol names that were renamed in this branch *)
		oldRenamings
	},
	
	(* take note of what has changed in lastCumulatedStatements in this branch: *)
	
	branchCumulatedStatements=Join[branchCumulatedStatements,
		Complement[lastCumulatedStatements,rememberedLastCumulatedStatements]
	];
	
	(* take note of what has changed in cumulatedAggregateAccesses in this branch: *)
	
	branchCumulatedAggregateAccesses=Join[branchCumulatedAggregateAccesses,
		Complement[cumulatedAggregateAccesses,rememberedCumulatedAggregateAccesses]
	];
	
	(* take note of new freezing renames in this branch: *)
	
	(* Note: in the following example:
		
		If[cond,
			x=...
		]
		
		there is a new renaming for symbol x within the conditional branch:
		
		Subscript[x, 1]\[Equal]...
		
		but since the branch is conditional, the final renaming after the branch is:
		
		cond => Subscript[x, 2]\[Equal]Subscript[x, 1]
		
		but this should also create a complementary renaming for the case when
		the condition is not met (that is, for the alternative branch, even if that
		alternative branch does not exist explicitly) :
		
		!cond => Subscript[x, 2]\[Equal]x
		
		so that after the whole If statement Subscript[x, 2] is properly bound to either
		Subscript[x, 1] (with new value assigned) or x (with the old value),
		depending on the branch condition.
		
		The above is implemented by creating a "strong" (overwriting) entry in
		conditionalFreezingUpdates for the cond case and "weak" (non-overwriting)
		entry for the !cond case (or, more generally, for all alternative conditions,
		which are listed in the otherBranchConditions parameter - there will be
		just one alternative branch for the If statement, but can be multiple
		with Switch statements that use multiple case labels), which allows
		to properly handle the case of missing alternative branches or the case of
		different branches creating renamings for disctinct or arbitrarily overlapping
		set of symbol names, for example:
		
		If[cond,
			x=...;   (* Subscript[x, 1] created here *)
			y=...,   (* Subscript[y, 2] created here *)
			
			y=...;   (* Subscript[y, 3] created here *)
			z=...    (* Subscript[z, 4] created here *)
		]
		
		The expected set of entries for this example is:
		
		cond  => Subscript[x, 5]\[Equal]Subscript[x, 1]
		cond  => Subscript[y, 6]\[Equal]Subscript[y, 2]
		cond  => Subscript[z, 7]\[Equal]z
		!cond => Subscript[x, 5]\[Equal]x
		!cond => Subscript[y, 6]\[Equal]Subscript[y, 3]
		!cond => Subscript[z, 7]\[Equal]Subscript[z, 4]
		
		and then the analysis can continue with Subscript[x, 5], Subscript[y, 6] and Subscript[z, 7].
	*)
	
	(* ensure container exists for the currently processed branch: *)
	If[Not[KeyExistsQ[conditionalFreezingUpdates,branchCondition]],
		conditionalFreezingUpdates[branchCondition]=<||>
	];
	(* and also ensure containers exist for all the other branches: *)
	Scan[
		Function[otherCond,
			If[Not[KeyExistsQ[conditionalFreezingUpdates,otherCond]],
				conditionalFreezingUpdates[otherCond]=<||>
			]
		],
		otherBranchConditions
	];
	
	(* Note: freezingRenames and rememberedFreezingRenames contain entries
		for all three objectT, objectParentT and objectLhsT wrapped names.
		This apparent duplication is convenient for creation of name replacement rules,
		but is of no benefit here and could actually confuse later creation of
		conditional renames.
		For this reason, structures used to handle conditional renames keep track
		for only one of these wrappers (objectT), they will be naturally replicated back
		when new conditional freezes are made.
	*)
	
	newRenamesInThisBranch=Complement[
		onlyObjectTKeyedEntries[freezingRenames],
		onlyObjectTKeyedEntries[rememberedFreezingRenames]
	];
	
	If[includeStrongRenamings,
		(* "strong" (overwriting) entries (note the order of Join): *)
		conditionalFreezingUpdates[branchCondition]=Join[
			conditionalFreezingUpdates[branchCondition],
			newRenamesInThisBranch
		]
	];
	
	If[includeWeakRenamings,
		(* check which symbols were overwritten by new renamings: *)
		
		newRenamedNames=Cases[Keys[newRenamesInThisBranch],_FMTSymbols`objectT,All];
		
		oldRenamings=Association@@Map[
			Function[oldName,
				oldName->Lookup[rememberedFreezingRenames,oldName,oldName]
			],
			newRenamedNames
		];
		
		(* "weak" (non-overwriting) entries in all other branches
			(note the order of Join - this ensures that existing entries are retained): *)
		Scan[
			Function[otherCond,
				conditionalFreezingUpdates[otherCond]=Join[
					oldRenamings,
					conditionalFreezingUpdates[otherCond]
				]
			],
			otherBranchConditions
		]
	]
]

SetAttributes[endBranch,HoldAll]


End[];


EndPackage[];
