(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


Needs["FMTSymbols`"];
Needs["FMTWorkarounds`"];


BeginPackage["FMTProver`"];


verifyProofs::usage="verifyProofs[sys,packageName,operationName,opProofStatements,variableDomains,enumMappings] Returns the result of verification of all proof conditions for the given operation.\n"<>
"The result has form {{index,description,wish,outcome},...}.";


Begin["Private`"];


(* Helper, proves a single set of wishes against their knowns,
	the proof is performed by finding instances for specific wishes;
	this helper is used for proving wishes wrapped in
	neverT (in which case the given condition is considered to fail),
	and possibleT (in which case the given condition passes as reachable).
*)
findExamples[knowns_,wishes_,variableDomains_]:=Module[
	{
		variables,
		domains, (* list {Element[var,domain],...}, only those which are used in knowns and wishes *)
		examples
	},
	
	variables=Union[
		Flatten[
			Map[
				Function[st,
					Cases[st,_Symbol|_Subscript|_FMTSymbols`initializedT,All]
				],
				Join[knowns,wishes]
			]
		]
	];
	
	domains=Map[
		Function[v,
			Element[v,variableDomains[v]]
		],
		variables
	];
	
	examples=FMTWorkarounds`findInstance[
		And@@Join[
			knowns,
			{And@@wishes},
			domains
		]
	];
	
	If[examples!={},
		First[examples],
		{}
	]
]


(* Helper, prooves a single set of wishes against their knowns,
	the proof is performed by finding instances for negated wishes;
	this helper is used for proving wishes wrapped in alwaysT
	(in which case the given condition is considered to fail).
*)
findContradictions[knowns_,wishes_,variableDomains_]:=
	findExamples[knowns,{Not[And@@wishes]},variableDomains]


(* Helper, reverts numeric mappings for enum types
	(that is, changes key\[Rule]value into value\[Rule]key mappings *)
revertMappings[mappings_]:=Map[
	Function[m, (* m is {k\[Rule]v,...} *)
		Association[Map[Reverse,m]]
	],
	mappings
]

(* Helper, applies reverse enum mappings to counterexamples, to recover original enumeration names
	on values found as solution instances in the Integers domain *)
applyReverseEnumMappings[singleResult_,originalDomains_,enumMappings_]:=With[
	{
		reverseEnumMappings=revertMappings[enumMappings]
	},
	
	Map[
		Function[ce, (* ce is name\[Rule]value *)
			With[
				{
					name=ce[[1]],
					value=ce[[2]]
				},
				
				If[KeyExistsQ[originalDomains,name],
					With[
						{
							domainOrType=originalDomains[name]
						},
						
						If[domainOrType=!=Integers&&domainOrType=!=Booleans,
							If[KeyExistsQ[reverseEnumMappings[domainOrType],value],
								name->reverseEnumMappings[domainOrType][value],
								name->Indeterminate
							],
							
							(* the type is Integers or Booleans, no restoration *)
							name->value
						]
					],
					
					(* value has no explicit domain assigned,
						(it is initializedT[var] tag, which is implicitly Boolean),
						\[Rule] no restoration *)
					name->value
				]
			]
		],
		singleResult
	]
]


(* Accepts the list of proof statements from FMTProofs`proofStatements
	(a single invocation is for a single operation)
	and returns the list {{index,{description,wish},outcome},...}, where
	index is a point number in the given operations body,
	description is a human-readable "correctness headline",
	wish is a statement that needs to be true in the given part of operation,
	outcome is one of:
		provenT[] (if the given wish is satisfied),
		impossibleT[] (if the possibleT wish is unreachable),
		failedT[contradictions] (if the given wish is not satisfied,
								contradictions is an example list of variable assignments (rules),
								for which the given wish fails)
*)
verifyProofs[sys_,
	packageName_String,
	qOperationName_Symbol,
	opProofStatements_,
	variableDomains_Association,
	enumMappings_
]:=Module[
	{
		mappedDomains,
		enumTypesUsed,
		enumReplacements,
		domains, (* domains for variables and their initialization flags *)
		singleProofResult,
		temp,
		results={} (* {{index,wish,outcome},...} *)
	},
	
	(* apply enumeration mappings to both proof statements and variable domains,
		but keep the original domain information to enable reverse mapping on found counterexamples *)
	
	enumTypesUsed=Keys[enumMappings];
	enumReplacements=Join@@Values[enumMappings];
	mappedDomains=variableDomains/.Thread[enumTypesUsed->Integers];
	
	(* in addition to all domains that were assigned for variables as used in operation body,
		assign also domains for tagged values like initializedT[var] *)
	
	temp=Cases[opProofStatements,
		in_FMTSymbols`initializedT->Rule[in,Booleans],
		All
	];
	temp=Union[temp];
	domains=Append[mappedDomains,temp];
	
	Scan[
		Function[proof, (* proof is {index,knowns,{{description,wish},...}} *)
			With[
				{
					index=proof[[1]],
					mappedKnowns=proof[[2]]/.enumReplacements,
					wishes=proof[[3]]
				},
				
				(* Note:
					for the given index, all wishes are processes separately;
					as an optimization strategy (to be checked!)
					consider grouping wishes of the same type
					(that is: group by neverT, alwaysT and possibleT)
				*)
				
				Scan[
					Function[describedWish, (* {description,wish} *)
						With[
							{
								description=describedWish[[1]],
								originalWish=describedWish[[2]],
								mappedWish=describedWish[[2]]/.enumReplacements
							},
							
							Switch[Head[originalWish],
								FMTSymbols`alwaysT,
									temp=findContradictions[
										mappedKnowns,
										List@@mappedWish,
										domains
									];
									If[temp!={},
										(* an example contradiction was found \[Rule]
											condition failed *)
										singleProofResult=applyReverseEnumMappings[
											FMTSymbols`failedT@@temp,variableDomains,enumMappings],
										
										(* no contradictions \[Rule] condition proved correct *)
										singleProofResult=FMTSymbols`provenT[]
									],
								
								FMTSymbols`neverT,
									temp=findExamples[
										mappedKnowns,
										List@@mappedWish,
										domains
									];
									If[temp!={},
										(* an example was found for what should never happen \[Rule]
											condition failed *)
										singleProofResult=applyReverseEnumMappings[
											FMTSymbols`failedT@@temp,variableDomains,enumMappings],
										
										(* no contradictions \[Rule] condition proved correct *)
										singleProofResult=FMTSymbols`provenT[]
									],
								
								FMTSymbols`possibleT,
									temp=findExamples[
										mappedKnowns,
										List@@mappedWish,
										domains
									];
									If[temp!={},
										(* an example was found \[Rule] condition proved correct *)
										singleProofResult=FMTSymbols`provenT[],
										
										(* no example found for what should be possible \[Rule]
											condition failed *)
										singleProofResult=FMTSymbols`impossibleT[]
									]
							];
							
							AppendTo[results,
								{
									index,
									{
										description,
										originalWish
									},
									singleProofResult
									
									(* Note: it might be possible to append this instead, as the last element of this list:
										
										singleProofResult/.reverseFreezings
										
										but then the same object can appear more than once in the list of solutions
										for a failing counterexample, which is misleading.
										By leaving the variable names in their frozen forms, it is also possible
										to track the evolution of the value from the beginning of the operation
										(where all object have their original symbol names) up to the point
										of proof (where objects have their frozen replacements).
										The price for this is a bit more cryptic report (with frozen names),
										but it seem to be worth the effort.
										
										The reverseFreezings values are returned as the last component of each element
										in the list from operationProofStatements, in case they are useful.
									*)
								}
							]
						]
					],
					wishes
				]
			]
		],
		opProofStatements
	];
	
	results
]


End[];


EndPackage[];
