(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


Needs["FMTGrammar`"];
Needs["FMTSymbols`"];
Needs["FMTSystem`"];


BeginPackage["FMTVerification`"];


operationConditions::usage="operationConditions[sys,qOperationName,resolvedBody,index] Returns the verification conditions for the given point (index) in the operation.\n"<>
"The return value has the form:\n"<>
"{{{description,pre-cond},...},{{description,post-cond},...},{known,...}}\n"<>
"where pre-cond has to be correct before the indexed statement is executed,\n"<>
"post-cond has to be correct after the indexed statement is executed,\n"<>
"known is a new fact statement on the given execution path\n"<>
"and description is a human-readable explanation of the condition in that pair.\n"<>
"Index can be start, finish or a number and any of the sublists in the result can be empty.";


Begin["Private`"];


(* Helper, verifies division by zero in the given expression *)
checkDivZero[conds_,expr_]:=Module[
	{
		temp
	},
	
	temp=Cases[expr,FMTSymbols`divT[x_,y_]->y,All];
	Scan[
		Function[y,
			AppendTo[conds,
				{
					"Division by zero.",
					FMTSymbols`neverT[y==0]
				}
			]
		],
		temp
	]
]

SetAttributes[checkDivZero,HoldFirst]


(* Helper, verifies array indices are within bounds in the given expression *)
checkArrayBounds[sys_,conds_,expr_]:=Module[
	{
		temp,
		qArrayType,
		bounds,
		idx
	},
	
	temp=Cases[expr,FMTSymbols`partT[a_,idx_],All];
	Scan[
		Function[p, (* p is partT[objectParentT[a,typeName],idx] *)
			qArrayType=p[[1,2]];
			bounds=FMTSystem`arrayTypeBounds[sys,qArrayType];
			idx=p[[2]];
			AppendTo[conds,
				{
					"Array index should be within bounds.",
					FMTSymbols`alwaysT[bounds[[1]]<=idx<=bounds[[2]]]
				}
			]
		],
		temp
	]
]

SetAttributes[checkArrayBounds,HoldAll]


(* Helper, verifies that those local variables that are used in the given expression
	(that is, those that are read in the given expression,
	which means that out parameters in operation calls are not included in this analysis)
	are already initialized *)
checkLocalVarsInitialized[sys_,qOperationName_,conds_,expr_]:=Module[
	{
		temp,
		localVarsThatNeedInitialization
	},
	
	(* variables that need to be initialized before use
		include all locals of the given operation except the in and inOut parameters,
		as they are considered to be already initialized by the caller:
	*)
	localVarsThatNeedInitialization=First/@FMTSystem`operationUninitializedLocals[sys,qOperationName];
	
	temp=Cases[expr,
		FMTSymbols`objectT[name_,type_]|FMTSymbols`objectParentT[name_,type_],
		All
	];
	Scan[
		Function[v, (* objectT[name,type] or objectParentT[name,type] *)
			If[MemberQ[localVarsThatNeedInitialization,First[v]],
				AppendTo[conds,
					{
						"Variable should be initialized before use.",
						FMTSymbols`alwaysT[FMTSymbols`initializedT[v]]
					}
				]
			]
		],
		temp
	]
]

SetAttributes[checkLocalVarsInitialized,HoldAll]


(* Helper, transforms an expression into a form
	where individual operations have at most two arguments *)
reduceOperatorArity[expr_]:=With[
	{
		replacements={
			FMTSymbols`plusT[a_,b_,rest__]->FMTSymbols`plusT[FMTSymbols`plusT[a,b],rest],
			FMTSymbols`timesT[a_,b_,rest__]->FMTSymbols`timesT[FMTSymbols`timesT[a,b],rest]
		}
	},
	
	FixedPoint[ReplaceAll[replacements],expr]
]

(* Helper, verifies a single subexpression against its base type *)
checkSubWithinBase[sys_,conds_,expr_]:=Module[
	{
		typeHier,
		enumValues,
		reducedExpr,
		qTypeName,
		baseType,
		baseInvariants
	},
	
	typeHier=FMTSystem`typeHierarchy[sys];
	enumValues=FMTSystem`allEnumValues[sys];
	reducedExpr=FMTGrammar`reduceAll[typeHier,enumValues,expr];
	
	(* reducedExpr is FMTSymbols`exprT[typeName] *)
	
	qTypeName=reducedExpr[[1]];
	baseType=qTypeName/.typeHier;
	
	baseInvariants=FMTSystem`typeInvariants[sys,baseType];
	
	AppendTo[conds,
		{
			"Subexpression should fit the base type.",
			FMTSymbols`alwaysT[baseInvariants/.baseType->expr]
		}
	]
]

SetAttributes[checkSubWithinBase,HoldAll]

(* Helper, verifies whether all subexpressions (including top-level) fit their base type *)
checkAllSubexpressionsWithinBase[sys_,conds_,expr_]:=Module[
	{
		reducedArity,
		reducedArityDepth,
		noLeafLevels,
		checkedOperators
	},
	
	reducedArity=reduceOperatorArity[expr];
	reducedArityDepth=Depth[reducedArity];
	
	(* all levels with root but without leaves,
		bottom-up (for readability in reports): *)
	noLeafLevels=Table[d,{d,reducedArityDepth-1,0,-1}];
	
	checkedOperators={
		FMTSymbols`divT,
		FMTSymbols`negT,
		FMTSymbols`plusT,
		FMTSymbols`timesT
	};
	
	Scan[
		Function[depth,
			Scan[
				Function[subEx,
					If[MemberQ[checkedOperators,Head[subEx]],
						checkSubWithinBase[sys,conds,subEx]
					]
				],
				reducedArity,
				{depth}
			]
		],
		noLeafLevels
	]
]

SetAttributes[checkAllSubexpressionsWithinBase,HoldAll]


(* Helper, verifies whether the given value fits the given type invariants *)
checkTypeInvariants[sys_,qType_,conds_,expr_]:=With[
	{
		invariants=FMTSystem`typeInvariants[sys,qType]
	},
	
	If[Length[invariants]>0,
		AppendTo[conds,
			{
				"Value should fit the target type.",
				FMTSymbols`alwaysT[invariants/.qType->expr]
			}
		]
	]
]

SetAttributes[checkTypeInvariants,HoldAll]


(* Helper, verifies whether in and inOut actual parameters are compatible
	with their respective formal parameter types *)
checkInputParametersTypeInvariants[sys_,conds_,formals_,actuals_]:=Module[
	{
		mode, (* parameter mode *)
		qType (* parameter's formal type *)
	},
	
	(* ScanThread would be better here, but it does not exist! *)
	MapThread[
		Function[{formal,actual}, (* formal is mode[name\[Rule]qType] *)
			mode=Head[formal];
			qType=formal[[1,2]];
			
			(* parameter type invariants are checked only for in and inOut modes: *)
			If[mode===FMTSymbols`in||mode===FMTSymbols`inOut,
				checkTypeInvariants[sys,qType,conds,actual]
			]
		],
		{List@@formals,List@@actuals}
	];
]

SetAttributes[checkInputParametersTypeInvariants,HoldAll]


checkCallPreconditions[sys_,packageName_,qCalledOpName_,conds_,formals_,actuals_]:=Module[
	{
		callPreConditions,
		paramReplacements
	},
	
	callPreConditions=FMTSystem`operationAssumptions[sys,packageName,qCalledOpName];
	If[callPreConditions!={},
		paramReplacements=MapThread[
			Function[{formal,actual}, (* formal is mode[name\[Rule]type] *)
				With[
					{
						formalName=formal[[1,1]]
					},
					
					formalName->actual
				]
			],
			{List@@formals,List@@actuals}
		];
		
		AppendTo[conds,
			{
				"Operation call pre-conditions.",
				FMTSymbols`alwaysT[And@@callPreConditions/.paramReplacements]
			}
		]
	]
]

SetAttributes[checkCallPreconditions,HoldAll]


registerCallPostconditions[sys_,packageName_,qCalledOpName_,knowns_,formals_,actuals_]:=Module[
	{
		callPostConditions,
		paramReplacements
	},
	
	callPostConditions=FMTSystem`operationPromises[sys,packageName,qCalledOpName];
	If[callPostConditions!={},
		paramReplacements=MapThread[
			Function[{formal,actual}, (* formal is mode[name\[Rule]type] *)
				With[
					{
						formalName=formal[[1,1]],
						formalType=formal[[1,2]]
					},
					
					(* post-conditions have grammar already resolved,
						which means that parameter names are already wrapped
						in objectT wrappers - this means that here the replacement
						rule needs to take the whole wrapper into account,
						not just the parameter name: *)
					FMTSymbols`objectT[formalName,formalType]->actual
				]
			],
			{List@@formals,List@@actuals}
		];
		
		AppendTo[knowns,
			And@@callPostConditions/.paramReplacements
		]
	]
]

SetAttributes[registerCallPostconditions,HoldAll]


(* Helper, verifies whether the conditional branch
	("then" in ifT or loop body in whileT) is reachable *)
checkReachable[conds_,expr_]:=
	AppendTo[conds,
		{
			"Branch is reachable.",
			FMTSymbols`possibleT[expr]
		}
	]

SetAttributes[checkReachable,HoldFirst]

(* Helper, verifies whether the alternative conditional branch
	("else" in ifT) is reachable *)
checkAlternativeReachable[conds_,expr_]:=
	AppendTo[conds,
		{
			"Alternative is reachable.",
			FMTSymbols`possibleT[Not[expr]]
		}
	]

SetAttributes[checkAlternativeReachable,HoldFirst]


(* Helper, verifies that local variables are used (initialized) on at least one existing path *)
checkLocalVarsReachable[sys_,qOperationName_,conds_]:=With[
	{
		localVars=FMTSystem`operationUninitializedLocals[sys,qOperationName]
	},
	
	Scan[
		Function[var, (* var is varName\[Rule]type *)
			AppendTo[conds,
				{
					"Variable should be initialized on some path.",
					FMTSymbols`possibleT[FMTSymbols`initializedT[FMTSymbols`objectT@@var]]
				}
			]
		],
		localVars
	]
]

SetAttributes[checkLocalVarsReachable,HoldAll]


(* Helper, verifies that operation's pre-conditions are reachable
	(that is, they do not conflict with invariants established as knowns
	at the very beginning of the operation, in the "first" index)
*)
registerAndCheckPreConditions[sys_,packageName_,qOperationName_,knowns_,conds_]:=With[
	{
		assumptions=FMTSystem`operationAssumptions[sys,packageName,qOperationName]
	},
	
	If[assumptions!={},
		knowns=Join[knowns,assumptions];
		AppendTo[conds,
			{
				"Operation pre-conditions are reachable.",
				FMTSymbols`possibleT[True]
			}
		]
	]
]

SetAttributes[registerAndCheckPreConditions,HoldAll]


(* Helper, verifies operation's post-conditions *)
checkPostConditions[sys_,packageName_,qOperationName_,conds_]:=With[
	{
		promises=FMTSystem`operationPromises[sys,packageName,qOperationName]
	},
	
	If[promises!={},
		AppendTo[conds,
			{
				"Operation post-conditions.",
				FMTSymbols`alwaysT@@promises
			}
		]
	]
]

SetAttributes[checkPostConditions,HoldAll]


(* Helper, verifies explicit user's assertions *)
checkAssert[conds_,expr_]:=
	AppendTo[conds,
		{
			"Explicit assertion.",
			If[MemberQ[{FMTSymbols`alwaysT,FMTSymbols`neverT,FMTSymbols`possibleT},Head[expr]],
				expr,
				FMTSymbols`alwaysT[expr]
			]
		}
	]

SetAttributes[checkAssert,HoldFirst]


(* Helpers for specific indices *)

startConditions[sys_,packageName_,qOperationName_,resolvedBody_]:=Module[
	{
		temp,
		postConds={},
		knowns={}
	},
	
	(* Note: there are no initializedT statements generated for variables
		that are known to be already initialized
		(these are params of mode in or inOut and package variables),
		and consistently there are no initializedT verification conditions generated
		for such variables;
		only uninitialized local variables have initializedT generated, both
		at the beginning of the operation and whenever they are accessed
	*)
	
	(* type invariants, even if implicit from type declarations,
		of all initialized objects used by the operation: *)
	temp=FMTSystem`operationInitializedObjects[sys,packageName,qOperationName,resolvedBody];
	Scan[
		Function[var, (* varName\[Rule]type *)
			With[
				{
					varName=var[[1]],
					varType=var[[2]],
					invariants=FMTSystem`typeInvariants[sys,var[[2]]]
				},
				
				AppendTo[knowns,invariants/.varType->varName]
			]
		],
		temp
	];
	
	(* Note: postConds is populated here, because pre-conditions
		should be checked against all knowns (which are invariants of all
		used variables) gathered at the "start" index,
		preConds would be checked before taking knowns into account,
		which would be pointless.
		Treating pre-conditions within postConds here does not change the fact
		that the whole check happens at the very beginning of the operation,
		before the first effective part of the operation's body.
	*)
	registerAndCheckPreConditions[sys,packageName,qOperationName,knowns,postConds];
	
	(* Not[initializedT[localVar]] for all uninitialized locals *)
	temp=FMTSystem`operationUninitializedLocals[sys,qOperationName];
	Scan[
		Function[var, (* varName\[Rule]type *)
			AppendTo[knowns,
				Not[FMTSymbols`initializedT[FMTSymbols`objectT@@var]]
			]
		],
		temp
	];
	
	(* no preConds here: *)
	{{},postConds,DeleteCases[knowns,True]}
]


finishConditions[sys_,packageName_,qOperationName_]:=Module[
	{
		postConds={}
	},
	
	(* "finish" is an artificial index, positioned after all other indices in the operation				-> it does not matter whether operation's final conditions are reported as pre- or
		post-conditions for that particular index, as there are no new knowns created
		in this point anyway and pre- and post-conditions are essentially equivalent;
		the particular choice of preConds/postConds sets below is not strict,
		and is only conventional to help readability *)
	
	(* all local vars can be initialized on some existing path
		(otherwise they would be dead code) *)
	checkLocalVarsReachable[sys,qOperationName,postConds];
	
	checkPostConditions[sys,packageName,qOperationName,postConds];
	
	(* no preConds and no knowns here: *)
	{{},postConds,{}}
]


assignmentConditions[sys_,qOperationName_,part_]:=Module[
	{
		localVars,
		object,
		qType,
		preConds={},
		postConds={},
		knowns={}
	},
	
	(* part is assignmentT[objectLhsT[obj,type],rhs] *)
	
	checkLocalVarsInitialized[sys,qOperationName,preConds,part[[2]]]; (* check in rhs *)
	Scan[
		Function[arrayIndex,
			(* check in lhs, within array index,
				if there is array access at the top level in lhs *)
			checkLocalVarsInitialized[sys,qOperationName,preConds,arrayIndex]
		],
		Cases[part[[1]],FMTSymbols`partT[_,idx_]->idx,{1}]
	];
	checkDivZero[preConds,part]; (* check in both lhs and rhs *)
	checkArrayBounds[sys,preConds,part]; (* check in both lhs and rhs *)
	
	(* all subexpressions need to fit their base type,
		but since the whole expression value will be checked on the lhs anyway,
		there is no need to analyze the whole expression with root here
		(hence withRoot parameter is False) *)
	
	checkAllSubexpressionsWithinBase[sys,preConds,part]; (* check both lhs and rhs *)
	
	(* the assigned value needs to fit the (invariants of) target type
		Note: invariants use type name as a placeholder for the actual value *)
	
	object=part[[1]];
	qType=part[[1,2]];
	
	checkTypeInvariants[sys,qType,postConds,object];
	
	(* assignment ensures that lhs is initialized,
		but this needs to be stated only for local variables: *)
	
	(* list of local variables (without their types): *)
	localVars=First/@FMTSystem`operationUninitializedLocals[sys,qOperationName];
	If[MemberQ[localVars,object[[1]]],
		(* assigned-to object happens to be a local variable: *)
		knowns={FMTSymbols`initializedT[object]}
	];
	
	{preConds,postConds,DeleteCases[knowns,True]}
]


callConditions[sys_,packageName_,qOperationName_,part_]:=Module[
	{
		qCalledOpName,
		localVars,
		inOrInOutActuals={},  (* not necessarily full-objects,
								for checking whether objects are already initialized *)
		inOutOrOutActuals={}, (* full-objects only, to register that they are initialized by a call *)
		preConds={},
		postConds={},
		knowns={}
	},
	
	(* part is callT[qOpName,formalParamsT[...],actualParamsT[...]] *)
	
	qCalledOpName=part[[1]];
	
	(* some check will apply to out actual parameters, while others to in and inOut
		- separate the parameters based on mode: *)
	MapThread[
		Function[{formal,actual},
			With[
				{
					mode=Head[formal],
					objName=actual[[1]]
				},
				
				(* to check whether objects used as in and inOut params are initialized,
					all uses are taken into account, including array element accesses
					and part dereferencing *)
				If[mode===FMTSymbols`in||mode===FMTSymbols`inOut,
					AppendTo[inOrInOutActuals,actual]
				];
				
				(* to register in knowns that objects used as inOut and out params
					are initialized by operation call,
					only full objects are taken into account *)
				If[Head[objName]===Symbol&&
					(mode===FMTSymbols`inOut||mode===FMTSymbols`out),
					
					AppendTo[inOutOrOutActuals,actual]
				]
			]
		],
		{List@@part[[2]],List@@part[[3]]}
	];
	
	checkLocalVarsInitialized[sys,qOperationName,preConds,inOrInOutActuals];
	checkDivZero[preConds,part[[3]]];
	checkArrayBounds[sys,preConds,part[[3]]];
	checkAllSubexpressionsWithinBase[sys,preConds,part[[3]]];
	checkInputParametersTypeInvariants[sys,preConds,part[[2]],part[[3]]];
	checkCallPreconditions[sys,packageName,qCalledOpName,preConds,part[[2]],part[[3]]];
	registerCallPostconditions[sys,packageName,qCalledOpName,knowns,part[[2]],part[[3]]];
	
	(* successful call ensures that out params are initialized,
		but this needs to be stated only for local variables,
		and only those that are used as whole objects: *)
	
	localVars=First/@FMTSystem`operationUninitializedLocals[sys,qOperationName];
	Scan[
		Function[object, (* object is objectT[name,type] *)
			If[MemberQ[localVars,object[[1]]],
				(* the object used as out param happens to be a local variable: *)
				AppendTo[knowns,FMTSymbols`initializedT[object]]
			]
		],
		inOutOrOutActuals
	];
	
	(* no postConds here: *)
	{preConds,{},DeleteCases[knowns,True]}
]


ifConditions[sys_,qOperationName_,part_]:=Module[
	{
		condition,
		preConds={}
	},
	
	(* the whole part at the given index is an If statement,
		check the controlling part: *)
	
	condition=part[[1]]; (* conditional expression *)
	checkLocalVarsInitialized[sys,qOperationName,preConds,condition];
	checkDivZero[preConds,condition];
	checkArrayBounds[sys,preConds,part];
	checkAllSubexpressionsWithinBase[sys,preConds,condition];
	checkReachable[preConds,condition];
	checkAlternativeReachable[preConds,condition];
	
	(* no postConds and no knowns here: *)
	{preConds,{},{}}
]


switchConditions[sys_,qOperationName_,part_]:=Module[
	{
		conditionObj,
		caseLabels,
		hasDefault,
		preConds={}
	},
	
	(* the whole part at the given index is a Switch statement,
		check the controlling part: *)
	
	conditionObj=part[[1]]; (* conditional enum object *)
	checkLocalVarsInitialized[sys,qOperationName,preConds,conditionObj];
	
	caseLabels=List@@part[[2;; ;;2]];
	caseLabels=caseLabels/.FMTSymbols`alternativesT->List;
	caseLabels=Flatten[caseLabels];
	hasDefault=MemberQ[caseLabels,_FMTSymbols`defaultT];
	caseLabels=DeleteCases[caseLabels,_FMTSymbols`defaultT];
	
	(* 1. check whether all explicitly named labels are reachable *)
	Scan[
		Function[cl,
			checkReachable[preConds,FMTSymbols`equalT[conditionObj,cl]]
		],
		caseLabels
	];
	
	(* 2. if there is a default label, check if it is possible
		to have value that is different from all explicitly mentioned *)
	If[hasDefault,
		checkReachable[preConds,
			And@@Map[
				Function[cl,
					FMTSymbols`unequalT[conditionObj,cl]
				],
				caseLabels
			]
		]
	];
	
	(* no postConds and no knowns here: *)
	{preConds,{},{}}
]


whileConditions[sys_,qOperationName_,part_]:=Module[
	{
		condition,
		preConds={}
	},
	
	(* the whole part at the given index is a While statement,
		check the controlling part: *)
	
	condition=part[[1]]; (* conditional expression *)
	checkLocalVarsInitialized[sys,qOperationName,preConds,condition];
	checkDivZero[preConds,condition];
	checkArrayBounds[sys,preConds,part];
	checkAllSubexpressionsWithinBase[sys,preConds,condition];
	checkReachable[preConds,condition];
	(* Note: alternative reachable should not be verified,
		because it is not a bug if the loop always executes at least once *)
	
	(* no postConds and no knowns here: *)
	{preConds,{},{}}
]


assertConditions[part_]:=Module[
	{
		preConds={}
	},

	(* part is assertT[...], or assertT[alwaysT[...]] or similar: *)
	checkAssert[preConds,part];
		
	(* no postConds and no knowns here: *)
	{preConds,{},{}}
]


assumeConditions[part_]:=Module[
	{
		knowns
	},

	(* explicit assumptions are added to knowns: *)
	(* part is assumeT[expr] *)					
	knowns={part}
		
	(* no preConds and no postConds here: *)
	{{},{},knowns}
]


operationConditions[sys_,packageName_,qOperationName_,resolvedBody_,index_]:=
	(* Note: conditions and assumptions can have some shared parts, for example from
		assertions, which are conditions for previous knowns and assumptions
		for the subsequent statements at the same time.
		
		Note: for dynamic checking, mixing of all conditions for the given index
		will not work, they need to be checked one-by-one, as the expression is evaluated.
	*)
	Switch[index,
		"start",
			startConditions[sys,packageName,qOperationName,resolvedBody],
		
		"finish",
			finishConditions[sys,packageName,qOperationName],
		
		_,
			(* regular index (numbered statement) *)
			With[
				{
					part=FMTSystem`operationPart[resolvedBody,index]
				},
				
				Switch[part,
					_FMTSymbols`assignmentT,
						assignmentConditions[sys,qOperationName,part],
					
					_FMTSymbols`callT,
						callConditions[sys,packageName,qOperationName,part],
					
					_FMTSymbols`ifT,
						ifConditions[sys,qOperationName,part],
					
					_FMTSymbols`switchT,
						switchConditions[sys,qOperationName,part],
					
					_FMTSymbols`whileT,
						whileConditions[sys,qOperationName,part],
					
					_FMTSymbols`assertT,
						assertConditions[part[[1]]],
					
					_FMTSymbols`assumeT,
						assumeConditins[part[[1]]]
				]
			]
	]


End[];


EndPackage[];
